"""ADRV9009 Utility Functions."""

from typing import Dict, Union

from adijif.converters.converter import converter


def _convert_to_config(
    L: Union[int, float],
    M: Union[int, float],
    Np: Union[int, float],
    S: Union[int, float] = 1,
) -> Dict:
    return {
        "L": L,
        "M": M,
        "F": Np / 8 * M * S / L,
        "S": S,
        "HD": (
            1
            if (M == 1 and L == 2 and S == 1)
            or (M == 2 and L == 4 and S == 1)
            or (M == 1 and L == 4 and S == 2)
            else 0
        ),
        "Np": Np,
        "N": Np,
        "CS": 0,
        "CF": 0,
        "K": 32,  # THIS IS A SUGGESTED VALUE
    }


quick_configuration_modes_tx = {
    # M = 1
    str(0): _convert_to_config(M=1, L=1, Np=16),
    str(1): _convert_to_config(M=1, L=2, Np=16),
    # M = 2
    str(2): _convert_to_config(M=2, L=1, Np=16),
    str(3): _convert_to_config(M=2, L=2, Np=16),
    str(4): _convert_to_config(M=2, L=4, Np=16),
    # M = 4
    str(5): _convert_to_config(M=4, L=1, Np=16),
    str(6): _convert_to_config(M=4, L=2, Np=16),
    str(7): _convert_to_config(M=4, L=4, Np=16),
    str(8): _convert_to_config(M=4, L=2, Np=12),
}


quick_configuration_modes_rx = {
    # M = 1
    str(0): _convert_to_config(M=1, L=1, S=1, Np=16),
    str(1): _convert_to_config(M=1, L=1, S=2, Np=16),
    str(2): _convert_to_config(M=1, L=1, S=4, Np=16),
    str(3): _convert_to_config(M=1, L=2, S=1, Np=16),
    str(4): _convert_to_config(M=1, L=2, S=2, Np=16),
    str(5): _convert_to_config(M=1, L=2, S=4, Np=16),
    str(6): _convert_to_config(M=1, L=4, S=2, Np=16),
    str(7): _convert_to_config(M=1, L=4, S=4, Np=16),
    # M = 2
    str(8): _convert_to_config(M=2, L=1, S=1, Np=16),
    str(9): _convert_to_config(M=2, L=1, S=2, Np=16),
    str(10): _convert_to_config(M=2, L=2, S=1, Np=16),
    str(11): _convert_to_config(M=2, L=2, S=2, Np=16),
    str(12): _convert_to_config(M=2, L=2, S=4, Np=16),
    str(13): _convert_to_config(M=2, L=4, S=1, Np=16),
    str(14): _convert_to_config(M=2, L=4, S=2, Np=16),
    str(15): _convert_to_config(M=2, L=4, S=4, Np=16),
    # M = 4
    str(16): _convert_to_config(M=4, L=1, S=1, Np=16),
    str(17): _convert_to_config(M=4, L=2, S=1, Np=16),
    str(18): _convert_to_config(M=4, L=2, S=2, Np=16),
    str(19): _convert_to_config(M=4, L=4, S=1, Np=16),
    str(20): _convert_to_config(M=4, L=4, S=2, Np=16),
    str(21): _convert_to_config(M=4, L=4, S=4, Np=16),
    # Np = 12,24
    str(22): _convert_to_config(M=2, L=1, S=2, Np=12),
    str(23): _convert_to_config(M=4, L=1, S=1, Np=12),
    str(24): _convert_to_config(M=4, L=2, S=1, Np=12),
    str(25): _convert_to_config(M=2, L=2, S=2, Np=24),
    str(26): _convert_to_config(M=4, L=2, S=1, Np=24),
}


def _extra_jesd_check(dev: converter) -> None:
    FK = dev.F * dev.K
    assert FK <= 256, "F x K must be <= 256"
    assert FK >= 20, "F x K must be >= 20"
    assert FK % 4 == 0, "F x K must be a multiple of 4"



"""Clock parent metaclass to maintain consistency for all clock chip."""

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.common import core
from adijif.draw import Layout, Node
from adijif.gekko_trans import gekko_translation


class clock(core, gekko_translation, metaclass=ABCMeta):
    """Parent metaclass for all clock chip classes."""

    @property
    @abstractmethod
    def find_dividers(self) -> Dict:
        """Find all possible divider settings that validate config.

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def list_available_references(self) -> List[int]:
        """Determine all references that can be generated.

        Based on config list possible references that can be generated
        based on VCO and output dividers

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError  # pragma: no cover

    def _solve_gekko(self) -> bool:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            bool: Always False
        """
        self.model.options.SOLVER = 1  # APOPT solver
        self.model.solver_options = [
            "minlp_maximum_iterations 1000",  # minlp iterations with integer solution
            "minlp_max_iter_with_int_sol 100",  # treat minlp as nlp
            "minlp_as_nlp 0",  # nlp sub-problem max iterations
            "nlp_maximum_iterations 500",  # 1 = depth first, 2 = breadth first
            "minlp_branch_method 1",  # maximum deviation from whole number
            "minlp_integer_tol 0",  # covergence tolerance (MUST BE 0 TFC)
            "minlp_gap_tol 0.1",
        ]

        self.model.solve(disp=False)
        self.model.cleanup()
        return False

    # def _add_objective(self, sysrefs: List) -> None:
    #     pass

    def _solve_cplex(self) -> CpoSolveResult:
        self.solution = self.model.solve(LogVerbosity="Quiet")
        if self.solution.solve_status not in ["Feasible", "Optimal"]:
            raise Exception("Solution Not Found")
        return self.solution

    def solve(self) -> Union[None, CpoSolveResult]:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            [None,CpoSolveResult]: When cplex solver is used CpoSolveResult is returned

        Raises:
            Exception: If solver is not valid

        """
        if self.solver == "gekko":
            return self._solve_gekko()
        elif self.solver == "CPLEX":
            return self._solve_cplex()
        else:
            raise Exception(f"Unknown solver {self.solver}")

    def draw(self, lo: Layout = None) -> str:
        """Generic Draw converter model.

        Args:
            lo (Layout): Layout object to draw on

        Returns:
            str: Path to image file

        Raises:
            Exception: If no solution is saved
        """
        if not self._saved_solution:
            raise Exception("No solution to draw. Must call solve first.")
        clocks = self._saved_solution
        system_draw = lo is not None
        name = self.name.lower()

        if not system_draw:
            lo = Layout(f"{name} Example")
        else:
            assert isinstance(lo, Layout), "lo must be a Layout object"

        ic_node = Node(self.name)
        lo.add_node(ic_node)

        # rate = clocks[f"{name}_ref_clk"]
        # Find key with ending
        ref_name = None
        for key in clocks.keys():
            if "vcxo" in key.lower():
                ref_name = key
                break
        if ref_name is None:
            raise Exception(f"No clock found for vcxo\n.Options: {clocks.keys()}")

        if not system_draw:
            ref_in = Node("REF_IN", ntype="input")
            lo.add_node(ref_in)
        else:
            to_node = lo.get_node(ref_name)
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            ref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

        rate = clocks[ref_name]

        lo.add_connection({"from": ref_in, "to": ic_node, "rate": rate})

        # Add each output clock
        for o_clk_name in clocks["output_clocks"]:
            rate = clocks["output_clocks"][o_clk_name]["rate"]
            # div = clocks['output_clocks'][o_clk_name]['divider']
            if not system_draw:
                out_node = Node(o_clk_name, ntype="out_clock_connected")
                lo.add_node(out_node)
                lo.add_connection({"from": ic_node, "to": out_node, "rate": rate})

        if not system_draw:
            return lo.draw()



"""AD9081 high speed MxFE clocking model."""

from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Union

import numpy as np

from ..solvers import GEKKO, CpoModel, CpoSolveResult  # type: ignore
from .ad9081_dp import ad9081_dp_rx, ad9081_dp_tx
from .ad9081_util import _load_rx_config_modes, _load_tx_config_modes
from .adc import adc
from .converter import converter
from .dac import dac

from ..solvers import integer_var  # type: ignore # isort: skip # noqa: I202


class ad9081_core(converter, metaclass=ABCMeta):
    """AD9081 high speed MxFE model.

    This model supports both direct clock configurations and on-board
    generation

    Clocking: AD9081 can internally generate or leverage external clocks. The
    high speed clock within the system is referred to as the DAC clock and
    the ADC clock will be a divided down version of the clock:
    adc_clock  == dac_clock / L, where L = 1,2,3,4


    For internal generation, the DAC clock is generated through an integer PLL
    through the following relation:
    dac_clock == ((m_vco * n_vco) / R * ref_clock) / D

    For external clocks, the clock must be provided at the DAC clock rate

    Once we have the DAC clock the data rates can be directly evaluated into
    each JESD framer:

    rx_baseband_sample_rate = (dac_clock / L) / datapath_decimation
    tx_baseband_sample_rate = dac_clock / datapath_interpolation

    """

    device_clock_available = None  # FIXME
    device_clock_ranges = None  # FIXME

    model: Union[GEKKO, CpoModel] = None

    name = "AD9081"

    # Integrated PLL constants
    l_available = [1, 2, 3, 4]
    l = 1  # pylint:  disable=E741
    m_vco_available = [5, 7, 8, 11]  # 8 is nominal
    m_vco = 8
    n_vco_available = [*range(2, 50 + 1)]
    n_vco = 2
    r_available = [1, 2, 3, 4]
    r = 1
    d_available = [1, 2, 3, 4]
    d = 1
    # Integrated PLL limits
    pfd_min = 25e6
    pfd_max = 750e6
    vco_min = 6e9
    vco_max = 12e9

    # JESD parameters
    available_jesd_modes = ["jesd204b", "jesd204c"]
    M_available = [1, 2, 3, 4, 6, 8, 12, 16]
    L_available = [1, 2, 3, 4, 6, 8]
    N_available = [12, 16]
    Np_available = [8, 12, 16, 24]
    F_available = [1, 2, 3, 4, 6, 8, 12, 16, 24, 32]
    S_available = [1, 2, 4, 8]
    # FIXME
    # K_available = [4, 8, 12, 16, 20, 24, 28, 32]
    K_available = [16, 32, 64, 128, 256]
    CS_available = [0, 1, 2, 3]
    CF_available = [0]
    # FIXME

    # Clocking constraints
    clocking_option_available = ["integrated_pll", "direct"]
    _clocking_option = "integrated_pll"
    bit_clock_min_available = {"jesd204b": 1.5e9, "jesd204c": 6e9}
    bit_clock_max_available = {"jesd204b": 15.5e9, "jesd204c": 24.75e9}

    config = {}  # type: ignore

    device_clock_max = 12e9
    _lmfc_divisor_sysref_available = [
        1,
        2,
        4,
        8,
        16,
        32,
        64,
        128,
        256,
        512,
        1024,
    ]

    def _check_valid_internal_configuration(self) -> None:
        # FIXME
        pass

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal converter configuration and output clock definitions
        leading to connected devices (clock chips, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        if solution:
            self.solution = solution
        if self.clocking_option == "integrated_pll":
            pll_config: Dict = {
                "m_vco": self._get_val(self.config["ad9081_m_vco"]),
                "n_vco": self._get_val(self.config["ad9081_n_vco"]),
                "r": self._get_val(self.config["ad9081_r"]),
                "d": self._get_val(self.config["ad9081_d"]),
            }
            if "serdes_pll_div" in self.config:
                pll_config["serdes_pll_div"] = self._get_val(
                    self.config["serdes_pll_div"]
                )
            return {
                "clocking_option": self.clocking_option,
                "pll_config": pll_config,
            }
        else:
            if "serdes_pll_div" in self.config:
                return {
                    "serdes_pll_div": self._get_val(self.config["serdes_pll_div"]),
                    "clocking_option": self.clocking_option,
                }
            return {"clocking_option": self.clocking_option}

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order
        """
        # clk = (
        # "ad9081_dac_clock" if self.clocking_option == "direct" else "ad9081_pll_ref"
        # )
        clk = f"{self.name.lower()}_ref_clk"
        sysref = f"{self.name.lower()}_sysref"
        return [clk, sysref]

    @property
    @abstractmethod
    def _converter_clock_config(self) -> None:
        """Define source clocking relation based on ADC, DAC, or both.

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError

    def _pll_config(self, rxtx: bool = False) -> Dict:
        self._converter_clock_config()  # type: ignore

        self.config["ad9081_m_vco"] = self._convert_input([5, 7, 8, 11], "ad9081_m_vco")
        self.config["ad9081_n_vco"] = self._convert_input(
            [*range(2, 51)], "ad9081_n_vco"
        )
        self.config["ad9081_r"] = self._convert_input([1, 2, 3, 4], "ad9081_r")
        self.config["ad9081_d"] = self._convert_input([1, 2, 3, 4], "ad9081_d")

        self.config["ad9081_ref_clk"] = self._add_intermediate(
            self.config["converter_clk"]
            * self.config["ad9081_d"]
            * self.config["ad9081_r"]
            / (self.config["ad9081_m_vco"] * self.config["ad9081_n_vco"])
        )
        # if self.solver == "gekko":
        #     self.config["ref_clk"] = self.model.Var(
        #         integer=True,
        #         lb=1e6,
        #         ub=self.device_clock_max,
        #         value=self.device_clock_max,
        #     )
        # elif self.solver == "CPLEX":
        #     # self.config["ref_clk"] = integer_var(
        #     #     int(1e6), int(self.device_clock_max), "ref_clk"
        #     # )
        #     self.config["ref_clk"] = (
        #         self.config["converter_clk"]
        #         * self.config["d"]
        #         * self.config["r"]
        #         / (self.config["m_vco"] * self.config["n_vco"])
        #     )
        # else:
        #     raise Exception("Unknown solver")

        self.config["ad9081_vco"] = self._add_intermediate(
            self.config["ad9081_ref_clk"]
            * self.config["ad9081_m_vco"]
            * self.config["ad9081_n_vco"]
            / self.config["ad9081_r"],
        )

        # if self.solver == "gekko":
        #     self.config["vco"] = self.model.Intermediate(
        #         self.config["ref_clk"]
        #         * self.config["m_vco"]
        #         * self.config["n_vco"]
        #         / self.config["r"],
        #     )
        # elif self.solver == "CPLEX":
        #     self.config["vco"] = (
        #         self.config["ref_clk"]
        #         * self.config["m_vco"]
        #         * self.config["n_vco"]
        #         / self.config["r"]
        #     )
        # else:
        #     raise Exception("Unknown solver: %s" % self.solver)

        self._add_equation(
            [
                self.config["ad9081_vco"] >= self.vco_min,
                self.config["ad9081_vco"] <= self.vco_max,
                self.config["ad9081_ref_clk"] / self.config["ad9081_r"] <= self.pfd_max,
                self.config["ad9081_ref_clk"] / self.config["ad9081_r"] >= self.pfd_min,
                self.config["ad9081_ref_clk"] >= int(100e6),
                self.config["ad9081_ref_clk"] <= int(2e9),
                # self.config["converter_clk"] <= self.device_clock_max,
                self.config["converter_clk"]
                >= (
                    self.converter_clock_min
                    if not rxtx
                    else self.dac.converter_clock_min  # type: ignore
                ),
                self.config["converter_clk"]
                <= (
                    self.converter_clock_max
                    if not rxtx
                    else self.dac.converter_clock_max  # type: ignore
                ),
            ]
        )

        # Make ref_clk an integer since API requires it
        if self.solver == "CPLEX":
            self.config["integer_ad9081_ref_clk"] = integer_var(
                min=int(100e6), max=int(2e9), name="integer_ad9081_ref_clk"
            )
            self._add_equation(
                [self.config["integer_ad9081_ref_clk"] == self.config["ad9081_ref_clk"]]
            )
        else:
            raise Exception("Only CPLEX solver supported")

        return self.config["ad9081_ref_clk"]

    def get_required_clocks(self) -> List:
        """Generate list required clocks.

        For AD9081 this will contain [converter clock, sysref requirement SOS]

        Returns:
            List: List of solver variables, equations, and constants
        """
        # SYSREF
        self.config = {}
        self.config["lmfc_divisor_sysref"] = self._convert_input(
            self._lmfc_divisor_sysref_available, "lmfc_divisor_sysref"
        )

        self.config["sysref"] = self._add_intermediate(
            self.multiframe_clock / self.config["lmfc_divisor_sysref"]
        )

        # Device Clocking
        if self.clocking_option == "direct":
            # raise Exception("Not implemented yet")
            clk = self.sample_clock * self.datapath_decimation
        else:
            clk = self._pll_config()  # type: ignore

        # Objectives
        # self.model.Obj(self.config["sysref"])  # This breaks many searches
        # self.model.Obj(-1*self.config["lmfc_divisor_sysref"])

        return [clk, self.config["sysref"]]


class ad9081_rx(adc, ad9081_core):
    """AD9081 Receive model."""

    name = "AD9081_RX"
    converter_type = "adc"

    converter_clock_min = 1.45e9
    converter_clock_max = 4e9

    sample_clock_min = 312.5e6 / 16
    sample_clock_max = 4e9

    quick_configuration_modes = _load_rx_config_modes()

    datapath = ad9081_dp_rx()
    decimation_available = [
        1,
        2,
        3,
        4,
        6,
        8,
        9,
        12,
        16,
        18,
        24,
        32,
        36,
        48,
        64,
        72,
        96,
        144,
        "auto",
    ]

    @property
    def decimation(self) -> int:
        """Decimation factor. This is the product of the coarse and fine decimation."""
        return self.datapath.decimation_overall

    @decimation.setter
    def decimation(self, value: int) -> None:
        raise Exception(
            "Decimation is not writable and should be set by the properties\n"
            + " datapath.cddc_decimations and datapath.fddc_decimations"
        )

    _adc_lmfc_divisor_sysref = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize AD9081 clocking model for RX.

        This is a common class used to handle RX constraints
        together.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        self.set_quick_configuration_mode("3.01", "jesd204b")
        self.datapath.cddc_decimations = [2] * 4
        self.datapath.fddc_decimations = [4] * 8
        self.datapath.fddc_enabled = [True] * 8
        assert self.decimation == 8
        super().__init__(*args, **kwargs)

    def _converter_clock_config(self) -> None:
        """RX specific configuration of internal PLL config.

        This method will update the config struct to include
        the RX clocking constraints
        """
        adc_clk = self.decimation * self.sample_clock
        self.config["l"] = self._convert_input([1, 2, 3, 4], "l")
        self.config["adc_clk"] = self._convert_input(adc_clk)
        self.config["converter_clk"] = self._add_intermediate(
            self.config["adc_clk"] * self.config["l"]
        )

    def _check_valid_internal_configuration(self) -> None:
        mode = self._check_valid_jesd_mode()
        cfg = self.quick_configuration_modes[self.jesd_class][mode]

        # Check decimation is valid
        if isinstance(self.decimation, int) or isinstance(self.decimation, float):
            found = False
            combos = []
            for dec in cfg["decimations"]:
                found = found or dec["coarse"] * dec["fine"] == self.decimation
                combos.append(f'{dec["coarse"]}/{dec["fine"]}')
            assert found, (
                f"Decimation {self.decimation} not valid for current JESD mode\n"
                + f"Valid CDDC/FDDC {combos}"
            )
        elif self.decimation == "auto":
            for dec in cfg["decimations"]:
                dec = dec["coarse"] * dec["fine"]
                # Check
                cc = dec * self.sample_clock
                # if dec == 64:
                #     print("dec", dec, cc, cfg["coarse"], cfg["fine"])
                if cc <= self.converter_clock_max and cc >= self.converter_clock_min:
                    self.decimation = dec
                    print("Decimation automatically determined:", dec)
                    return
            raise Exception("No valid decimation found")
        else:
            raise Exception("Decimation not valid")


class ad9081_tx(dac, ad9081_core):
    """AD9081 Transmit model."""

    name = "AD9081_TX"
    converter_type = "dac"

    converter_clock_min = 2.9e9
    converter_clock_max = 12e9

    sample_clock_min = 2.9e9 / (6 * 24)  # with max interpolation
    sample_clock_max = 12e9

    quick_configuration_modes = _load_tx_config_modes()

    datapath = ad9081_dp_tx()
    interpolation_available = [
        1,
        2,
        3,
        4,
        6,
        8,
        9,
        12,
        16,
        18,
        24,
        32,
        36,
        48,
        64,
        72,
        96,
        144,
    ]

    @property
    def interpolation(self) -> int:
        """Interpolation factor.

        This is the product of the coarse and fine interpolation.

        Returns:
            int: Interpolation factor
        """
        return self.datapath.interpolation_overall

    @interpolation.setter
    def interpolation(self, value: int) -> None:
        raise Exception(
            "Interpolation is not writable and should be set by the properties\n"
            + " datapath.cduc_interpolation and datapath.fduc_interpolation"
        )

    _dac_lmfc_divisor_sysref = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize AD9081 clocking model for TX.

        This is a common class used to handle TX constraints
        together.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        self.set_quick_configuration_mode("0", "jesd204c")
        self.datapath.cduc_interpolation = 6
        self.datapath.fduc_interpolation = 4
        self.datapath.fduc_enabled = [True] * 8
        super().__init__(*args, **kwargs)

    def _converter_clock_config(self) -> None:
        """TX specific configuration of internall PLL config.

        This method will update the config struct to include
        the TX clocking constraints
        """
        dac_clk = self.interpolation * self.sample_clock
        self.config["dac_clk"] = self._convert_input(dac_clk)
        self.config["converter_clk"] = self._add_intermediate(self.config["dac_clk"])


class ad9081(ad9081_core):
    """AD9081 combined transmit and receive model."""

    converter_clock_min = ad9081_rx.converter_clock_min
    converter_clock_max = ad9081_rx.converter_clock_max
    quick_configuration_modes: Dict[str, Any] = {}
    _nested = ["adc", "dac"]
    converter_type = "adc_dac"

    def __init__(
        self, model: Union[GEKKO, CpoModel] = None, solver: str = None
    ) -> None:
        """Initialize AD9081 clocking model for TX and RX.

        This is a common class used to handle TX and RX constraints
        together.

        Args:
            model (GEKKO,CpoModel): Solver model
            solver (str): Solver name (gekko or CPLEX)
        """
        if solver:
            self.solver = solver
        self.adc = ad9081_rx(model, solver=self.solver)
        self.dac = ad9081_tx(model, solver=self.solver)
        self.model = model

    def validate_config(self) -> None:
        """Validate device configurations including JESD and clocks of both ADC and DAC.

        This check only is for static configuration that does not include
        variables which are solved.
        """
        self.adc.validate_config()
        self.dac.validate_config()

    def _get_converters(self) -> List[Union[converter, converter]]:
        return [self.adc, self.dac]

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order
        """
        clk = (
            "ad9081_dac_clock"
            if self.adc.clocking_option == "direct"
            else "ad9081_pll_ref"
        )
        return [clk, "ad9081_adc_sysref", "ad9081_dac_sysref"]

    def _converter_clock_config(self) -> None:
        adc_clk = self.adc.decimation * self.adc.sample_clock
        dac_clk = self.dac.interpolation * self.dac.sample_clock
        l = dac_clk / adc_clk
        if np.abs(l - round(l)) > 1e-6:
            raise Exception(f"Sample clock ratio is not integer {adc_clk} {dac_clk}")
        else:
            l = int(round(l))
        if l not in self.adc.l_available:
            raise Exception(
                f"ADC clock must be DAC clock/L where L={self.adc.l_available}."
                + f" Got {l} ({dac_clk}/{adc_clk})"
            )

        self.config["dac_clk"] = self._convert_input(dac_clk)
        self.config["adc_clk"] = self._convert_input(adc_clk)
        self.config["converter_clk"] = self._add_intermediate(self.config["dac_clk"])

        # Add single PLL constraint
        # JESD204B/C transmitter is a power of 2 divisor of the lane rate of
        # the JESD204B/C receiver
        if self.solver == "gekko":
            raise Exception("Not implemented for GEKKO")
        elif self.solver == "CPLEX":
            divs = [int(2**d) for d in range(16)]
            self.config["serdes_pll_div"] = self._convert_input(
                divs, "serdes_pll_div", default=1
            )
        else:
            raise Exception(f"Unknown solver {self.solver}")

        self._add_equation(
            [self.config["serdes_pll_div"] * self.adc.bit_clock == self.dac.bit_clock]
        )

    def get_required_clocks(self) -> List:
        """Generate list required clocks.

        For AD9081 this will contain [converter clock, sysref requirement SOS]

        Returns:
            List: List of solver variables, equations, and constants
        """
        # SYSREF
        self.config = {}
        self.config["adc_lmfc_divisor_sysref"] = self._convert_input(
            self.adc._adc_lmfc_divisor_sysref, "adc_lmfc_divisor_sysref"
        )
        self.config["dac_lmfc_divisor_sysref"] = self._convert_input(
            self.dac._dac_lmfc_divisor_sysref, "dac_lmfc_divisor_sysref"
        )

        self.config["sysref_adc"] = self._add_intermediate(
            self.adc.multiframe_clock / self.config["adc_lmfc_divisor_sysref"]
        )
        self.config["sysref_dac"] = self._add_intermediate(
            self.dac.multiframe_clock / self.config["dac_lmfc_divisor_sysref"]
        )

        # Device Clocking
        if self.clocking_option == "direct":
            # raise Exception("Not implemented yet")
            # adc_clk = self.sample_clock * self.datapath_decimation
            # clk = dac_clk
            clk = self.dac.interpolation * self.dac.sample_clock
        else:
            clk = self._pll_config(rxtx=True)

        # Objectives
        # self.model.Obj(self.config["sysref"])  # This breaks many searches
        # self.model.Obj(-1*self.config["lmfc_divisor_sysref"])

        return [clk, self.config["sysref_adc"], self.config["sysref_dac"]]


class ad9082_rx(ad9081_rx):
    """AD9082 MxFE RX Clocking Model."""

    converter_clock_max = 6e9


class ad9082_tx(ad9081_tx):
    """AD9082 MxFE TX Clocking Model."""

    pass


class ad9082(ad9081):
    """AD9081 combined transmit and receive model."""

    converter_clock_min = ad9082_rx.converter_clock_min
    converter_clock_max = ad9082_rx.converter_clock_max
    quick_configuration_modes: Dict[str, Any] = {}
    _nested = ["adc", "dac"]

    def __init__(
        self, model: Union[GEKKO, CpoModel] = None, solver: str = None
    ) -> None:
        """Initialize AD9081 clocking model for TX and RX.

        This is a common class used to handle TX and RX constraints
        together.

        Args:
            model (GEKKO,CpoModel): Solver model
            solver (str): Solver name (gekko or CPLEX)
        """
        if solver:
            self.solver = solver
        self.adc = ad9082_rx(model, solver=self.solver)
        self.dac = ad9082_tx(model, solver=self.solver)
        self.model = model

    def _get_converters(self) -> List[Union[converter, converter]]:
        return [self.adc, self.dac]

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order
        """
        clk = (
            "ad9082_dac_clock"
            if self.adc.clocking_option == "direct"
            else "ad9082_pll_ref"
        )
        return [clk, "ad9082_adc_sysref", "ad9082_dac_sysref"]



"""LTC6953 clock chip model."""

from typing import Dict, List, Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.clocks.clock import clock
from adijif.solvers import CpoExpr, GK_Intermediate


class ltc6953(clock):
    """LTC6953 clock chip model.

    This model currently supports all divider configurations
    """

    name = "LTC6953"

    input_ref = 1000000000

    # Ranges
    m_divider_min = 1
    m_divider_max = 1023
    m_available = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        52,
        54,
        56,
        58,
        60,
        62,
        64,
        68,
        72,
        76,
        80,
        84,
        88,
        92,
        96,
        100,
        104,
        108,
        112,
        116,
        120,
        124,
        128,
        136,
        144,
        152,
        160,
        168,
        176,
        184,
        192,
        200,
        208,
        216,
        224,
        232,
        240,
        248,
        256,
        272,
        288,
        304,
        320,
        336,
        352,
        368,
        384,
        400,
        416,
        432,
        448,
        464,
        480,
        496,
        512,
        544,
        576,
        608,
        640,
        672,
        704,
        736,
        768,
        800,
        832,
        864,
        896,
        928,
        960,
        992,
        1024,
        1088,
        1152,
        1216,
        1280,
        1344,
        1408,
        1472,
        1536,
        1600,
        1664,
        1728,
        1792,
        1856,
        1920,
        1984,
        2048,
        2176,
        2304,
        2432,
        2560,
        2688,
        2816,
        2944,
        3072,
        3200,
        3328,
        3456,
        3584,
        3712,
        3840,
        3968,
        4096,
    ]

    """ Output dividers """

    # Defaults
    _m: Union[int, List[int]] = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        52,
        54,
        56,
        58,
        60,
        62,
        64,
        68,
        72,
        76,
        80,
        84,
        88,
        92,
        96,
        100,
        104,
        108,
        112,
        116,
        120,
        124,
        128,
        136,
        144,
        152,
        160,
        168,
        176,
        184,
        192,
        200,
        208,
        216,
        224,
        232,
        240,
        248,
        256,
        272,
        288,
        304,
        320,
        336,
        352,
        368,
        384,
        400,
        416,
        432,
        448,
        464,
        480,
        496,
        512,
        544,
        576,
        608,
        640,
        672,
        704,
        736,
        768,
        800,
        832,
        864,
        896,
        928,
        960,
        992,
        1024,
        1088,
        1152,
        1216,
        1280,
        1344,
        1408,
        1472,
        1536,
        1600,
        1664,
        1728,
        1792,
        1856,
        1920,
        1984,
        2048,
        2176,
        2304,
        2432,
        2560,
        2688,
        2816,
        2944,
        3072,
        3200,
        3328,
        3456,
        3584,
        3712,
        3840,
        3968,
        4096,
    ]

    # Limits
    """ Internal limits """
    input_freq_max = 4.5e9

    # State management
    _clk_names: List[str] = []

    @property
    def m(self) -> Union[int, List[int]]:
        """Output dividers.

        Valid dividers are 1,2,3,4,5,6..32->(even)->4096

        Returns:
            int: Current allowable dividers
        """
        return self._d

    @m.setter
    def m(self, value: Union[int, List[int]]) -> None:
        """Output dividers.

        Valid dividers are 1,2,3,4,5,6..32->(even)->4096

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.m_available, "d")
        self._d = value

    def find_dividers(self) -> Dict:
        """Find the best dividers for the current configuration.

        NOT IMPLEMENTED YET

        Raises:
            NotImplementedError: Not implemented yet
        """
        raise NotImplementedError("find_dividers not implemented")

    def list_available_references(self) -> List[int]:
        """List the available reference frequencies.

        NOT IMPLEMENTED YET

        Raises:
            NotImplementedError: Not implemented yet
        """
        raise NotImplementedError("list_available_references not implemented")

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        out_dividers = [self._get_val(x) for x in self.config["out_dividers"]]

        config: Dict = {
            "out_dividers": out_dividers,
            "output_clocks": [],
        }

        # Handle different input_ref types
        if hasattr(self, "input_ref_arb") and self.input_ref_arb:
            # arb_source case
            input_ref_cfg = self.input_ref_arb.get_config(self.solution)  # type: ignore
            config["input_ref"] = list(input_ref_cfg.values())[0]
        else:
            # int/float or range case
            config["input_ref"] = self._get_val(
                self.input_ref
            )  # pytype: disable=attribute-error

        output_cfg = {}
        for i, div in enumerate(out_dividers):
            rate = config["input_ref"] / div  # type: ignore # noqa: B950
            output_cfg[self._clk_names[i]] = {"rate": rate, "divider": div}

        config["output_clocks"] = output_cfg

        self._saved_solution = config

        return config

    def _setup_solver_constraints(self, input_ref: int) -> None:
        """Apply constraints to solver model.

        Args:
            input_ref (int): Input reference frequency in hertz
        """
        self.config = {}
        if not isinstance(input_ref, (int, float)):
            input_ref_result = input_ref(self.model)  # type: ignore
            # Handle range type (returns dict with "range" key)
            if isinstance(input_ref_result, dict):
                self.config["input_ref_set"] = input_ref_result
                self.input_ref_arb = None
                input_ref = self.config["input_ref_set"]["range"]
            # Handle arb_source type (returns direct expression)
            else:
                self.config["input_ref_set"] = None
                self.input_ref_arb = input_ref  # Store original for get_config
                input_ref = input_ref_result
        else:
            self.config["input_ref_set"] = None
            self.input_ref_arb = None
        self.input_ref = input_ref

        self._add_equation(
            [
                input_ref <= self.input_freq_max,
            ]
        )

    def _setup(self, input_ref: int) -> None:
        if isinstance(input_ref, (float, int)):
            assert self.input_freq_max >= input_ref >= 0, "Input frequency out of range"

        # Setup clock chip internal constraints
        self._setup_solver_constraints(input_ref)

        # Add requested clocks to output constraints
        self.config["out_dividers"] = []

    def _get_clock_constraint(
        self, clk_name: List[str]
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  String of clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference

        Raises:
            Exception: Invalid solver
        """
        if self.solver == "gekko":
            __m = self._m if isinstance(self._m, list) else [self._m]
            if __m.sort() != self.m_available.sort():
                raise Exception("For solver gekko d is not configurable for LTC6952")
            mp = self.model.Var(integer=True, lb=1, ub=32)
            nx = self.model.Var(integer=True, lb=0, ub=7)
            od = self.model.Intermediate(mp * pow(2, nx))
        elif self.solver == "CPLEX":
            od = self._convert_input(self._m, f"m_{clk_name}")
        else:
            raise Exception("Unknown solver {}".format(self.solver))
        self.config["out_dividers"].append(od)
        return self.input_ref / od

    def set_requested_clocks(
        self, input_ref: int, out_freqs: List, clk_names: List[str]
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            input_ref (int): Input reference frequency in hertz
            out_freqs (List): list of required clocks to be output
            clk_names (List[str]):  list of strings of clock names

        Raises:
            Exception: If len(out_freqs) != len(clk_names)
        """
        if len(clk_names) != len(out_freqs):
            raise Exception("clk_names is not the same size as out_freqs")
        self._clk_names = clk_names

        # Setup clock chip internal constraints
        self._setup(input_ref)

        # Add requested clocks to output constraints
        for out_freq, clk_name in zip(out_freqs, clk_names):  # noqa: B905
            if self.solver == "gekko":
                __m = self._d if isinstance(self.__m, list) else [self.__m]
                if __m.sort() != self.m_available.sort():
                    raise Exception(
                        "For solver gekko m is not configurable for LTC6953"
                    )

                mp = self.model.Var(integer=True, lb=1, ub=32)
                nx = self.model.Var(integer=True, lb=0, ub=7)
                od = self.model.Intermediate(mp * pow(2, nx))

            elif self.solver == "CPLEX":
                od = self._convert_input(self._m, f"m_{out_freq}_{clk_name}")

            self._add_equation([self.input_ref / od == out_freq])
            self.config["out_dividers"].append(od)



import numpy as np

from adijif.converters.dac import dac


class ad9144_bf(dac):
    """Brute force methods for calculating clocks

    These are currently meant for debug to compare against
    the solver solutions
    """

    def device_clock_available(self):
        """Generate list of possible device clocks"""
        raise Exception("Not implemented")
        # aicd = sorted(self.available_input_clock_dividers)

        # dev_clocks = []
        # for div in aicd:
        #     in_clock = self.sample_clock * self.datapath_decimation * div
        #     if in_clock <= self.max_input_clock:
        #         dev_clocks.append(in_clock)
        # if not dev_clocks:
        #     raise Exception(
        #         "No device clocks possible in current config. Sample rate too high"
        #     )
        # return dev_clocks

    def device_clock_ranges(self):
        """Generate min and max values for device clock"""

        clks = self.device_clock_available()
        return np.min(clks), np.max(clks)

    def sysref_clock_ranges(self):
        """sysref must be multiple of LMFC"""
        lmfc = self.multiframe_clock
        return lmfc / 2048, lmfc / 2

    def sysref_met(self, sysref_clock, sample_clock):
        if sysref_clock % self.multiframe_clock != 0:
            raise Exception("SYSREF not a multiple of LMFC")
        if (self.multiframe_clock / sysref_clock) < 2 * self.input_clock_divider:
            raise Exception("SYSREF not a multiple of LMFC > 1")



# flake8: noqa
import pprint

import pytest

import adijif


@pytest.mark.parametrize("attr", ["data_path_width", "K", "F", "L", "M", "N", "Np"])
def test_jesd_ints(attr):
    with pytest.raises(Exception, match=f"{attr} must be an integer"):
        cnv = adijif.ad9680()
        setattr(cnv, attr, 0.5)


@pytest.mark.parametrize("attr", ["K", "F", "L", "M", "N", "Np"])
def test_jesd_oor(attr):
    with pytest.raises(Exception, match=f"{attr} not in range for device"):
        cnv = adijif.ad9680()
        setattr(cnv, attr, 1024)


def test_jesd_hd_parameter():
    """Test HD (high density) parameter."""
    cnv = adijif.ad9680()

    # Test setting HD
    cnv.HD = 1
    assert cnv.HD == 1

    cnv.HD = 0
    assert cnv.HD == 0


def test_jesd_cs_parameter():
    """Test CS (control bits) parameter."""
    cnv = adijif.ad9680()

    # Test setting CS
    cnv.CS = 0
    assert cnv.CS == 0


def test_jesd_s_parameter():
    """Test S (samples per converter per frame) parameter."""
    cnv = adijif.ad9680()

    # Test setting S
    cnv.S = 1
    assert cnv.S == 1


def test_jesd_jesd_class():
    """Test jesd_class property."""
    cnv = adijif.ad9680()

    # Test setting jesd_class
    cnv.jesd_class = "jesd204b"
    assert cnv.jesd_class == "jesd204b"

    # cnv.jesd_class = "jesd204c"
    # assert cnv.jesd_class == "jesd204c"


def test_jesd_multiframe_clock():
    """Test multiframe_clock calculation."""
    cnv = adijif.ad9680()

    cnv.sample_clock = 1e9
    cnv.decimation = 1
    cnv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Multiframe clock should be calculated
    mfc = cnv.multiframe_clock
    assert mfc > 0
    assert isinstance(mfc, (int, float))


def test_jesd_bit_clock():
    """Test bit_clock calculation."""
    cnv = adijif.ad9680()

    cnv.sample_clock = 1e9
    cnv.decimation = 1
    cnv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Bit clock should be calculated
    bc = cnv.bit_clock
    assert bc > 0
    assert isinstance(bc, (int, float))


def test_jesd_sample_clock():
    """Test sample_clock calculation."""
    cnv = adijif.ad9680()

    cnv.sample_clock = 1e9
    assert cnv.sample_clock == 1e9


def test_jesd_device_clock():
    """Test device_clock calculation."""
    cnv = adijif.ad9680()

    cnv.sample_clock = 1e9
    cnv.decimation = 1
    cnv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Device clock should be calculated
    dc = cnv.device_clock
    assert dc > 0


def test_jesd_lane_rate():
    """Test lane_rate calculation."""
    cnv = adijif.ad9680()

    cnv.sample_clock = 1e9
    cnv.decimation = 1
    cnv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Lane rate should be calculated
    lr = cnv.bit_clock
    assert lr > 0



"""Top-level package for pyadi-jif."""

__author__ = """Analog Devices, Inc."""
__email__ = "travis.collins@analog.com"
__version__ = "0.1.2"

import adijif.solvers
import adijif.utils
from adijif.clocks.ad9523 import ad9523_1
from adijif.clocks.ad9528 import ad9528
from adijif.clocks.ad9545 import ad9545
from adijif.clocks.hmc7044 import hmc7044
from adijif.clocks.ltc6952 import ltc6952
from adijif.clocks.ltc6953 import ltc6953
from adijif.converters.ad9081 import (
    ad9081,
    ad9081_rx,
    ad9081_tx,
    ad9082,
    ad9082_rx,
    ad9082_tx,
)
from adijif.converters.ad9084 import ad9084_rx, ad9088_rx
from adijif.converters.ad9144 import ad9144
from adijif.converters.ad9680 import ad9680
from adijif.converters.adrv9009 import adrv9009, adrv9009_rx, adrv9009_tx
from adijif.fpgas.xilinx import xilinx
from adijif.plls.adf4030 import adf4030
from adijif.plls.adf4371 import adf4371
from adijif.plls.adf4382 import adf4382
from adijif.system import system
from adijif.types import range



# flake8: noqa
import pprint

import pytest

import adijif

# solvers_to_test = ["gekko", "CPLEX"]
solvers_to_test = ["CPLEX"]


@pytest.mark.parametrize("solver", solvers_to_test)
def test_smoke_solver(solver):
    vcxo = 125000000
    sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo, solver=solver)
    sys.fpga.setup_by_dev_kit_name("zc706")

    sys.converter.sample_clock = 1e9 / 2
    sys.converter.datapath_decimation = 1
    sys.converter.L = 4
    sys.converter.M = 2
    sys.converter.N = 14
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 1

    sys.solve()


@pytest.mark.parametrize("solver", ["CPLEX"])
@pytest.mark.parametrize("converter", ["ad9680", "ad9144"])
@pytest.mark.parametrize("clockchip", ["ad9528", "hmc7044", "ad9523_1"])
@pytest.mark.parametrize("fpga_kit", ["zc706", "zcu102"])
def test_smoke_all_clocks(solver, converter, clockchip, fpga_kit):
    vcxo = 125000000
    sys = adijif.system(converter, clockchip, "xilinx", vcxo, solver=solver)
    sys.fpga.setup_by_dev_kit_name(fpga_kit)

    if converter == "ad9680":
        sys.converter.sample_clock = 1e9 / 2
        sys.converter.datapath_decimation = 1
        sys.converter.L = 4
        sys.converter.M = 2
        sys.converter.N = 14
        sys.converter.Np = 16
        sys.converter.K = 32
        sys.converter.F = 1
        sys.converter.HD = 1
    elif converter == "ad9144":
        sys.converter.sample_clock = 1e9
        # Mode 0
        sys.converter.interpolation = 1
        sys.converter.L = 8
        sys.converter.M = 4
        sys.converter.N = 16
        sys.converter.Np = 16
        sys.converter.K = 32
        sys.converter.F = 1
        sys.converter.HD = 1
        sys.converter.clocking_option = "integrated_pll"
    else:
        raise Exception("Unknown converter")

    sys.solve()


@pytest.mark.parametrize("solver", solvers_to_test)
@pytest.mark.parametrize(
    "qpll, cpll, rate", [(0, 0, 1e9), (0, 1, 1e9 / 2), (1, 0, 1e9 / 2)]
)
@pytest.mark.parametrize("clock_chip", ["ad9523_1", "hmc7044", "ad9528"])
@pytest.mark.parametrize("fpga_kit", ["zc706", "zcu102"])
def test_ad9680_all_clk_chips_fpga_pll_modes_solver(
    qpll, cpll, rate, clock_chip, solver, fpga_kit
):
    if fpga_kit == "zcu102" and clock_chip == "hmc7044" and rate == 1e9:
        pytest.skip()

    vcxo = 125000000

    sys = adijif.system("ad9680", clock_chip, "xilinx", vcxo, solver=solver)

    # Get Converter clocking requirements
    sys.converter.sample_clock = rate
    sys.converter.datapath_decimation = 1
    sys.converter.L = 4
    sys.converter.M = 2
    sys.converter.N = 14
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 1

    sys.fpga.setup_by_dev_kit_name(fpga_kit)
    sys.fpga.force_cpll = cpll
    sys.fpga.force_qpll = qpll
    sys.fpga.out_clk_select = "XCVR_REFCLK"

    if solver == "gekko":
        sys.fpga.favor_cpll_over_qpll = True
        sys.fpga.minimize_fpga_ref_clock = True
        sys.clock.minimize_feedback_dividers = False

    try:
        o = sys.solve()
    except Exception as e:
        sys.model = []
        del sys
        raise e
    sys.model = []
    del sys

    # Check
    if qpll:
        assert o["fpga_AD9680"]["type"] == "qpll"
    elif cpll:
        assert o["fpga_AD9680"]["type"] == "cpll"


@pytest.mark.parametrize("solver", solvers_to_test)
def test_daq2_split_rates_solver(solver):
    vcxo = 125000000
    sys = adijif.system(["ad9680", "ad9144"], "ad9523_1", "xilinx", vcxo, solver=solver)
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.out_clk_select = "XCVR_REFCLK"

    # Get Converter clocking requirements
    sys.converter[0].sample_clock = 1e9 / 2
    sys.converter[0].decimation = 1
    sys.converter[0].L = 4
    sys.converter[0].M = 2
    sys.converter[0].N = 14
    sys.converter[0].Np = 16
    sys.converter[0].K = 32
    sys.converter[0].F = 1
    sys.converter[0].HD = 1

    sys.converter[1].sample_clock = 1e9
    sys.converter[1].interpolation = 1
    sys.converter[1].L = 4
    sys.converter[1].M = 2
    sys.converter[1].N = 16
    sys.converter[1].Np = 16
    sys.converter[1].K = 32
    sys.converter[1].F = 1
    sys.converter[1].HD = 1

    cfg = sys.solve()

    assert cfg["fpga_AD9680"]["type"] == "cpll"  # CPLL
    assert cfg["fpga_AD9144"]["type"] == "qpll"  # QPLL
    assert cfg["fpga_AD9144"]["sys_clk_select"] == "XCVR_QPLL0"


def test_ad9680_clock_check1_solver():
    vcxo = 125000000
    sys = adijif.system("ad9680", "ad9523_1", "xilinx", vcxo)

    sys.fpga.out_clk_select = "XCVR_REFCLK"

    sys.converter.sample_clock = 1e9

    sys.fpga.setup_by_dev_kit_name("zc706")
    print(sys.fpga.out_clk_select)
    cfg = sys.solve()

    assert cfg["clock"]["n2"] == 24
    assert cfg["clock"]["r2"] == 1
    assert cfg["clock"]["m1"] == 3
    assert cfg["clock"]["output_clocks"]["zc706_AD9680_ref_clk"]["rate"] == 1e9 / 4
    assert cfg["fpga_AD9680"]["type"] == "qpll"
    assert cfg["fpga_AD9680"]["sys_clk_select"] == "XCVR_QPLL0"


@pytest.mark.xfail(reason="Need to verify settings")
@pytest.mark.parametrize("solver", solvers_to_test)
def test_ad9680_clock_check2_solver(solver):
    vcxo = 125000000

    sys = adijif.system("ad9680", "ad9523_1", "xilinx", vcxo, solver=solver)

    # Get Converter clocking requirements
    sys.converter.sample_clock = 1e9
    sys.converter.datapath_decimation = 1
    sys.converter.L = 4
    sys.converter.M = 2
    sys.converter.N = 14
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 1

    # Get FPGA clocking requirements
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.out_clk_select = "XCVR_REFCLK"

    cfg = sys.solve()
    print(cfg)

    clk_config = cfg["clock"]

    # Check clock math for AD9523-1
    assert vcxo / clk_config["r2"] * clk_config["n2"] >= sys.clock.vco_min
    assert vcxo / clk_config["r2"] * clk_config["n2"] <= sys.clock.vco_max

    divs = cfg["clock"]["out_dividers"]
    assert cfg["clock"]["n2"] == 24
    assert cfg["clock"]["r2"] == 1
    assert cfg["clock"]["m1"] == 3
    assert cfg["clock"]["output_clocks"]["AD9680_fpga_ref_clk"]["rate"] == 250000000
    for div in divs:
        assert div in [1, 4, 32]



# flake8: noqa

import pytest

import adijif


def test_converter_lane_count_valid():
    sys = adijif.system("ad9144", "ad9523_1", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    sys.converter.sample_clock = 1e9
    # Mode 0
    sys.converter.interpolation = 1
    sys.converter.L = 8
    sys.converter.M = 4
    sys.converter.N = 16
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 1
    sys.converter.clocking_option = "integrated_pll"

    cfg = sys.solve()


def test_converter_lane_count_exceeds_fpga_lane_count():
    convs = ["ad9144", "ad9144", "ad9144"]
    sys = adijif.system(convs, "ad9523_1", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    for i, _ in enumerate(convs):
        # Mode 0
        sys.converter[i].sample_clock = 1e9
        sys.converter[i].interpolation = 1
        sys.converter[i].L = 8
        sys.converter[i].M = 4
        sys.converter[i].N = 16
        sys.converter[i].Np = 16
        sys.converter[i].K = 32
        sys.converter[i].F = 1
        sys.converter[i].HD = 1
        sys.converter[i].clocking_option = "integrated_pll"

    with pytest.raises(Exception, match=f"Max SERDES lanes exceeded. 8 only available"):
        cfg = sys.solve()


def test_nested_converter_lane_count_valid():
    sys = adijif.system("adrv9009", "ad9528", "xilinx", 122.88e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.converter.adc.sample_clock = 122.88e6
    sys.converter.dac.sample_clock = 122.88e6

    sys.converter.adc.decimation = 4
    sys.converter.dac.interpolation = 4

    mode_rx = "17"
    mode_tx = "6"
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")

    cfg = sys.solve()


def test_nested_converter_lane_count_exceeds_fpga_lane_count():
    fpga_L = 1

    sys = adijif.system("adrv9009", "ad9528", "xilinx", 122.88e6)

    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.fpga.max_serdes_lanes = fpga_L  # Force it to break

    sys.converter.adc.sample_clock = 122.88e6
    sys.converter.dac.sample_clock = 122.88e6

    sys.converter.adc.decimation = 4
    sys.converter.dac.interpolation = 4

    mode_rx = "17"
    mode_tx = "6"
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")

    with pytest.raises(
        Exception, match=f"Max SERDES lanes exceeded. {fpga_L} only available"
    ):
        cfg = sys.solve()


def test_system_with_arb_source_vcxo():
    """Test system with arb_source VCXO type."""
    import adijif

    vcxo = adijif.types.arb_source("vcxo")

    sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo)
    sys.converter.sample_clock = 1e9
    sys.converter.decimation = 1
    sys.converter.set_quick_configuration_mode(str(0x88))
    sys.converter.K = 32

    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.force_qpll = 1

    cfg = sys.solve()
    assert cfg is not None


def test_system_model_reset():
    """Test that model_reset properly resets all components."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)

    # Store original model
    original_model = sys.model

    # Reset the model
    sys._model_reset()

    # Model should be recreated
    assert sys.model is not original_model
    assert sys.clock.model is sys.model
    assert sys.fpga.model is sys.model
    assert sys.converter.model is sys.model


def test_system_debug_solver():
    """Test system with debug solver enabled."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.Debug_Solver = True

    sys.converter.sample_clock = 1e9
    sys.converter.decimation = 1
    sys.converter.set_quick_configuration_mode(str(0x88))
    sys.converter.K = 32

    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.force_qpll = 1

    cfg = sys.solve()
    assert cfg is not None


def test_system_disabled_clocks():
    """Test system with both converter and FPGA clocks disabled."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.enable_converter_clocks = False
    sys.enable_fpga_clocks = False

    with pytest.raises(Exception, match="Converter and/or FPGA clocks must be enabled"):
        sys.solve()


def test_system_with_pll_inline():
    """Test system with external PLL inline setup."""
    import adijif

    sys = adijif.system("ad9144", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    sys.converter.sample_clock = 1e9
    sys.converter.interpolation = 1
    sys.converter.L = 8
    sys.converter.M = 4
    sys.converter.N = 16
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 1
    sys.converter.clocking_option = "integrated_pll"

    # Just test that add_pll_inline works
    sys.add_pll_inline("adf4382", sys.clock, sys.converter)
    assert len(sys.plls) == 1


def test_system_unknown_converter_type():
    """Test system with converter having invalid type."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.converter.sample_clock = 1e9
    sys.converter.decimation = 1
    sys.converter.set_quick_configuration_mode(str(0x88))
    sys.converter.K = 32

    # Force an invalid converter type
    sys.converter.converter_type = "INVALID"

    sys.fpga.setup_by_dev_kit_name("zc706")

    with pytest.raises(Exception, match="Unknown converter type"):
        sys.solve()


def test_system_use_common_sysref_attribute():
    """Test system with common sysref attribute."""
    import adijif

    sys = adijif.system("ad9144", "ad9523_1", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    # Test that attribute can be set
    sys.use_common_sysref = True
    assert sys.use_common_sysref is True


def test_system_arb_source_with_gekko_error():
    """Test that arb_source with gekko solver raises error."""
    import adijif

    vcxo = adijif.types.arb_source("vcxo")

    with pytest.raises(
        Exception,
        match="arb_source type requires CPLEX solver",
    ):
        sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo, solver="gekko")


def test_system_model_reset_with_list_converters():
    """Test model reset with list of converters."""
    import adijif

    convs = ["ad9144", "ad9144"]
    sys = adijif.system(convs, "ad9523_1", "xilinx", 125e6)

    original_model = sys.model

    sys._model_reset()

    assert sys.model is not original_model
    for conv in sys.converter:
        assert conv.model is sys.model


def test_system_destructor():
    """Test system destructor cleanup."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)

    sys.__del__()

    assert sys.fpga == []
    assert sys.converter == []
    assert sys.clock == []


def test_system_destructor_with_list():
    """Test system destructor with list of converters."""
    import adijif

    convs = ["ad9144", "ad9144"]
    sys = adijif.system(convs, "ad9523_1", "xilinx", 125e6)

    sys.__del__()

    assert sys.converter == []


def test_system_with_gekko_solver():
    """Test system with GEKKO solver."""
    import adijif

    try:
        sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6, solver="gekko")
        assert sys.solver == "gekko"

        # Test model reset with gekko
        sys._model_reset()
        assert sys.model is not None
    except Exception as e:
        # GEKKO may not be installed
        if "GEKKO Solver not installed" not in str(e):
            raise


def test_system_unknown_solver():
    """Test system with unknown solver."""
    import adijif

    with pytest.raises(Exception, match="Unknown solver"):
        sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6, solver="INVALID")


def test_system_model_reset_with_gekko():
    """Test model reset with GEKKO solver."""
    import adijif

    try:
        sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6, solver="gekko")
        original_model = sys.model

        sys._model_reset()

        assert sys.model is not original_model
        assert sys.clock.model is sys.model
    except Exception as e:
        # GEKKO may not be installed
        if "GEKKO Solver not installed" not in str(e):
            raise


@pytest.mark.skipif(True, reason="Old BF method, can take a long time to run")
def test_system_determine_clocks():
    """Test determine_clocks method."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.converter.sample_clock = 500e6
    sys.converter.decimation = 1
    sys.converter.set_quick_configuration_mode(str(0x88))
    sys.converter.K = 32

    sys.fpga.setup_by_dev_kit_name("zc706")

    try:
        configs = sys.determine_clocks()
        assert configs is not None
        assert len(configs) > 0
    except Exception:
        # May not find valid config for all parameter combinations
        pass


def test_system_plls_property():
    """Test plls property."""
    import adijif

    sys = adijif.system("ad9144", "hmc7044", "xilinx", 125e6)

    # Initially should be empty
    assert sys.plls == []
    assert len(sys.plls) == 0


def test_system_add_pll_inline_sets_connected_output():
    """Test that add_pll_inline properly sets connection."""
    import adijif

    sys = adijif.system("ad9144", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    sys.add_pll_inline("adf4382", sys.clock, sys.converter)

    assert len(sys._plls) == 1
    assert sys._plls[0]._connected_to_output == sys.converter.name


def test_system_solve_with_gekko():
    """Test solve with GEKKO solver."""
    import adijif

    try:
        sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6, solver="gekko")
        sys.converter.sample_clock = 500e6
        sys.converter.decimation = 1
        sys.converter.set_quick_configuration_mode(str(0x88))
        sys.converter.K = 32

        sys.fpga.setup_by_dev_kit_name("zc706")

        cfg = sys.solve()
        assert cfg is not None
    except Exception as e:
        # GEKKO may not be installed or may not find solution
        if "GEKKO Solver not installed" not in str(e):
            # Solution not found is acceptable
            pass


def test_system_solve_unknown_solver():
    """Test solve with unknown solver raises error."""
    import adijif

    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.solver = "INVALID_SOLVER"

    sys.converter.sample_clock = 1e9
    sys.converter.decimation = 1
    sys.converter.set_quick_configuration_mode(str(0x88))
    sys.converter.K = 32

    sys.fpga.setup_by_dev_kit_name("zc706")

    with pytest.raises(Exception, match="Unknown solver"):
        sys.solve()



"""AD9523-1 clock chip model."""

from typing import Dict, List, Union

from adijif.clocks.ad9523_1_bf import ad9523_1_bf
from adijif.solvers import CpoExpr, CpoIntVar, CpoSolveResult, GK_Intermediate


class ad9523_1(ad9523_1_bf):
    """AD9523-1 clock chip model.

    This model currently supports VCXO+PLL2 configurations
    """

    name = "AD9523-1"

    # Ranges
    m1_available = [3, 4, 5]
    d_available = [*range(1, 1024)]
    n2_available = [12, 16, 17, 20, 21, 22, 24, 25, 26, *range(28, 255)]
    a_available = [*range(0, 4)]
    b_available = [*range(3, 64)]
    # N = (PxB) + A, P=4, A==[0,1,2,3], B=[3..63]
    # See table 46 of DS for limits
    r2_available = list(range(1, 31 + 1))

    # Defaults
    _m1: Union[List[int], int] = [3, 4, 5]
    _d: Union[List[int], int] = [*range(1, 1024)]
    _n2: Union[List[int], int] = [
        12,
        16,
        17,
        20,
        21,
        22,
        24,
        25,
        26,
        *range(28, 255),
    ]
    _r2: Union[List[int], int] = list(range(1, 31 + 1))

    # Limits
    vco_min = 2.94e9
    vco_max = 3.1e9
    pfd_max = 259e6

    # State management
    _clk_names: List[str] = []

    minimize_feedback_dividers = True

    """ Enable internal VCXO/PLL1 doubler """
    use_vcxo_double = False

    vcxo: Union[int, float, CpoIntVar] = 125e6

    @property
    def m1(self) -> Union[int, List[int]]:
        """VCO divider path 1.

        Valid dividers are 3,4,5

        Returns:
            int: Current allowable dividers
        """
        return self._m1

    @m1.setter
    def m1(self, value: Union[int, List[int]]) -> None:
        """VCO divider path 1.

        Valid dividers are 3,4,5

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.m1_available, "m1")
        self._m1 = value

    @property
    def d(self) -> Union[int, List[int]]:
        """Output dividers.

        Valid dividers are 1->1023

        Returns:
            int: Current allowable dividers
        """
        return self._d

    @d.setter
    def d(self, value: Union[int, List[int]]) -> None:
        """Output dividers.

        Valid dividers are 1->1023

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.d_available, "d")
        self._d = value

    @property
    def n2(self) -> Union[int, List[int]]:
        """n2: VCO feedback divider.

        Valid dividers are 12, 16, 17, 20, 21, 22, 24, 25, 26, 28->255

        Returns:
            int: Current allowable dividers
        """
        return self._n2

    @n2.setter
    def n2(self, value: Union[int, List[int]]) -> None:
        """VCO feedback divider.

        Valid dividers are 12, 16, 17, 20, 21, 22, 24, 25, 26, 28->255

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.n2_available, "n2")
        self._n2 = value

    @property
    def r2(self) -> Union[int, List[int]]:
        """VCXO input dividers.

        Valid dividers are 1->31

        Returns:
            int: Current allowable dividers
        """
        return self._r2

    @r2.setter
    def r2(self, value: Union[int, List[int]]) -> None:
        """VCXO input dividers.

        Valid dividers are 1->31

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.r2_available, "r2")
        self._r2 = value

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")
        for k in ["out_dividers", "m1", "n2", "r2"]:
            if k not in self.config.keys():
                raise Exception("Missing key: " + str(k))

        if solution:  # type: ignore
            self.solution = solution
        config: Dict = {
            "m1": self._get_val(self.config["m1"]),
            "n2": self._get_val(self.config["n2"]),
            "r2": self._get_val(self.config["r2"]),
            "out_dividers": [self._get_val(x) for x in self.config["out_dividers"]],
            "output_clocks": [],
        }

        # Handle different vcxo types
        if hasattr(self, "vcxo_arb") and self.vcxo_arb:
            # arb_source case
            vcxo_cfg = self.vcxo_arb.get_config(self.solution)  # type: ignore
            config["vcxo"] = list(vcxo_cfg.values())[0]
        else:
            # int/float or range case
            config["vcxo"] = self._get_val(self.vcxo)  # pytype: disable=attribute-error
        vcxo = config["vcxo"]

        clk = vcxo / config["r2"] * config["n2"] / config["m1"]
        output_cfg = {}
        for i, div in enumerate(self.config["out_dividers"]):
            div = self._get_val(div)
            rate = clk / div
            output_cfg[self._clk_names[i]] = {"rate": rate, "divider": div}
        config["output_clocks"] = output_cfg
        config["vco"] = clk
        config["part"] = "AD9523-1"

        self._saved_solution = config

        return config

    def _setup_solver_constraints(self, vcxo: Union[float, int, CpoIntVar]) -> None:
        """Apply constraints to solver model.

        Args:
            vcxo (int): VCXO frequency in hertz

        Raises:
            Exception: Unknown solver
        """
        self.config = {
            "r2": self._convert_input(self._r2, "r2"),
            "m1": self._convert_input(self._m1, "m1"),
            "n2": self._convert_input(self._n2, "n2"),
        }
        if not isinstance(vcxo, (int, float)):
            vcxo_result = vcxo(self.model)  # type: ignore
            # Handle range type (returns dict with "range" key)
            if isinstance(vcxo_result, dict):
                self.config["vcxo_set"] = vcxo_result
                self.vcxo_arb = None
                vcxo = self.config["vcxo_set"]["range"]
            # Handle arb_source type (returns direct expression)
            else:
                self.vcxo_arb = vcxo  # Store original for get_config
                self.config["vcxo_set"] = None
                vcxo = vcxo_result
        else:
            self.vcxo_arb = None
            self.config["vcxo_set"] = None
        self.vcxo = vcxo

        # PLL2 equations
        self._add_equation(
            [
                vcxo / self.config["r2"] <= self.pfd_max,
                vcxo / self.config["r2"] * self.config["n2"] <= self.vco_max,
                vcxo / self.config["r2"] * self.config["n2"] >= self.vco_min,
            ]
        )
        # Objectives
        if self.minimize_feedback_dividers:
            if self.solver == "CPLEX":
                ...
                # self.model.minimize(self.config["n2"])
                # cost = self.model
                # self.model.minimize_static_lex([self.config["n2"],])
            elif self.solver == "gekko":
                self.model.Obj(self.config["n2"])
            else:
                raise Exception("Unknown solver {}".format(self.solver))

    def _add_objective(self, sys_refs: List[CpoIntVar]) -> None:
        # Minimize feedback divider and sysref frequencies
        if self.minimize_feedback_dividers:
            self._objectives = [self.config["n2"]] + sys_refs
            # self.model.add(
            #     self.model.minimize_static_lex([self.config["n2"]] + sys_refs)
            # )
        else:
            self._objectives = [sys_refs]
            # self.model.add(self.model.minimize_static_lex(sys_refs))

    def _setup(self, vcxo: int) -> None:
        # Setup clock chip internal constraints

        # FIXME: ADD SPLIT m1 configuration support

        if self.use_vcxo_double and not isinstance(vcxo, int):
            raise Exception("VCXO doubler not supported in this mode TBD")
        if self.use_vcxo_double:
            vcxo *= 2
        self._setup_solver_constraints(vcxo)
        self.config["out_dividers"] = []
        self._clk_names = []  # Reset clock names to be generated

    def _get_clock_constraint(
        self, clk_name: List[str]
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  String of clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference
        """
        od = self._convert_input(self._d, "d_" + str(clk_name))
        self.config["out_dividers"].append(od)
        self._clk_names.append(clk_name)
        return (
            self.vcxo / self.config["r2"] * self.config["n2"] / self.config["m1"] / od
        )

    def set_requested_clocks(
        self, vcxo: int, out_freqs: List, clk_names: List[str]
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            vcxo (int): VCXO frequency in hertz
            out_freqs (List): list of required clocks to be output
            clk_names (List[str]):  list of strings of clock names

        Raises:
            Exception: If len(out_freqs) != len(clk_names)
        """
        if len(clk_names) != len(out_freqs):
            raise Exception("clk_names is not the same size as out_freqs")

        # Setup clock chip internal constraints
        self._setup(vcxo)
        self._clk_names = clk_names

        # Add requested clocks to output constraints
        for out_freq in out_freqs:
            # od = self.model.Var(integer=True, lb=1, ub=1023, value=1)
            od = self._convert_input(self._d, "d_" + str(out_freq))
            # od = self.model.sos1([n*n for n in range(1,9)])

            self._add_equation(
                [
                    self.vcxo
                    / self.config["r2"]
                    * self.config["n2"]
                    / self.config["m1"]
                    / od
                    == out_freq
                ]
            )
            self.config["out_dividers"].append(od)



"""7 series transceiver model."""

from typing import List, Union

from docplex.cp.modeler import if_then

from ...converters.converter import converter as conv
from ...solvers import CpoIntVar, GK_Intermediate, GK_Operators, GKVariable
from .pll import PLLCommon, XilinxPLL


class SevenSeries(XilinxPLL):
    """7 series Transceiver model."""

    # References
    # GTXs
    # https://docs.amd.com/v/u/en-US/ug476_7Series_Transceivers
    # https://docs.amd.com/v/u/en-US/ds191-XC7Z030-XC7Z045-data-sheet

    transceiver_types_available = [
        "GTXE2",
        "GTHE2",
    ]  # We don't support GTHE2 yet!
    _transceiver_type = "GTXE2"

    force_cpll = False
    force_qpll = False

    def add_plls(self) -> None:
        """Add PLLs to the model."""
        self.plls = {"CPLL": CPLL(self), "QPLL": QPLL(self)}

    def add_constraints(
        self,
        config: dict,
        fpga_ref: Union[CpoIntVar, GK_Intermediate, GK_Operators, GKVariable, int],
        converter: conv,
    ) -> dict:
        """Add constraints for PLLs.

        Args:
            config (dict): Configuration dictionary.
            fpga_ref (int, CpoIntVar): FPGA reference clock.
            converter (conv): Converter object.

        Returns:
            dict: Updated configuration dictionary.
        """
        assert self.plls, "No PLLs configured. Run the add_plls method"
        assert not (self.force_cpll and self.force_qpll), "Both CPLL and QPLL enabled"
        for pll in self.plls:
            config = self.plls[pll].add_constraints(config, fpga_ref, converter)
        self._add_equation(
            config[converter.name + "_use_cpll"] + config[converter.name + "_use_qpll"]
            == 1
        )
        return config

    def get_config(
        self, config: dict, converter: conv, fpga_ref: Union[int, float]
    ) -> dict:
        """Get PLL configuration.

        Args:
            config (dict): Configuration dictionary.
            converter (conv): Converter object.
            fpga_ref (Union[int, float]): FPGA reference clock.

        Returns:
            dict: Updated configuration dictionary.
        """
        if self.force_cpll:
            ecpll = 1
            eqpll = self.solution.get_kpis()[converter.name + "_use_qpll"]
        elif self.force_qpll:
            ecpll = self.solution.get_kpis()[converter.name + "_use_cpll"]
            eqpll = 1
        else:
            ecpll = self.solution.get_kpis()[converter.name + "_use_cpll"]
            eqpll = self.solution.get_kpis()[converter.name + "_use_qpll"]

        assert ecpll != eqpll, "Both CPLL and QPLL enabled"
        pll = "CPLL" if ecpll else "QPLL"
        return self.plls[pll].get_config(config, converter, fpga_ref)


class CPLL(PLLCommon):
    """Channel PLL (CPLL) for 7 series FPGAs."""

    @property
    def vco_min(self) -> int:
        """Get the minimum VCO frequency for the transceiver type.

        Returns:
            int: Minimum VCO frequency.

        Raises:
            Exception: Unsupported transceiver type.
        """
        if self.parent.transceiver_type == "GTXE2":
            return 1600000000
        elif self.parent.transceiver_type == "GTHE2":
            return 1600000000
        raise Exception(
            f"Unknown vco_min for transceiver type {self.parent.transceiver_type}"
        )

    @property
    def vco_max(self) -> int:
        """Get the maximum VCO frequency for the transceiver type.

        Returns:
            int: Maximum VCO frequency.

        Raises:
            Exception: Unsupported transceiver type.
        """
        if self.parent.transceiver_type == "GTXE2":
            return 3300000000
        elif self.parent.transceiver_type == "GTHE2":
            return 5160000000
        raise Exception(
            f"Unknown vco_max for transceiver type {self.parent.transceiver_type}"
        )

    M_available = [1, 2]
    _M = [1, 2]

    @property
    def M(self) -> Union[int, List[int]]:
        """Get the M value for the CPLL."""
        return self._M

    @M.setter
    def M(self, value: Union[int, List[int]]) -> None:
        """Set the M value for the CPLL."""
        self._check_in_range(value, self.M_available, "M")
        self._M = value

    N2_available = [1, 2, 3, 4, 5]
    _N2 = [1, 2, 3, 4, 5]

    @property
    def N2(self) -> Union[int, List[int]]:
        """Get the N2 value for the CPLL."""
        return self._N2

    @N2.setter
    def N2(self, value: Union[int, List[int]]) -> None:
        """Set the N2 value for the CPLL."""
        self._check_in_range(value, self.N2_available, "N2")
        self._N2 = value

    N1_available = [4, 5]
    _N1 = [4, 5]

    @property
    def N1(self) -> Union[int, List[int]]:
        """Get the N1 value for the CPLL."""
        return self._N1

    @N1.setter
    def N1(self, value: Union[int, List[int]]) -> None:
        """Set the N1 value for the CPLL."""
        self._check_in_range(value, self.N1_available, "N1")
        self._N1 = value

    D_available = [1, 2, 4, 8]
    _D = [1, 2, 4, 8]

    @property
    def D(self) -> Union[int, List[int]]:
        """Get the D value for the CPLL."""
        return self._D

    @D.setter
    def D(self, value: Union[int, List[int]]) -> None:
        """Set the D value for the CPLL."""
        self._check_in_range(value, self.D_available, "D")
        self._D = value

    def get_config(
        self, config: dict, converter: conv, fpga_ref: Union[int, CpoIntVar]
    ) -> dict:
        """Get CPLL configuration.

        Args:
            config (dict): Configuration dictionary.
            converter (conv): Converter object.
            fpga_ref (int, CpoIntVar): FPGA reference clock.

        Returns:
            dict: Updated configuration dictionary.
        """
        pll_config = {}
        pll_config["type"] = "cpll"
        for k in ["m", "d", "n1", "n2"]:
            pll_config[k] = self._get_val(config[converter.name + "_" + k + "_cpll"])

        pll_config["vco"] = (
            fpga_ref * pll_config["n1"] * pll_config["n2"] / pll_config["m"]  # type: ignore # noqa: B950
        )
        # Check
        assert (
            pll_config["vco"] * 2 / pll_config["d"] == converter.bit_clock  # type: ignore # noqa: B950
        ), "Invalid CPLL lane rate"

        return pll_config

    def add_constraints(
        self, config: dict, fpga_ref: Union[int, CpoIntVar], converter: conv
    ) -> dict:
        """Add Channel PLL (CPLL) constraints.

        Args:
            config (dict): Configuration dictionary.
            fpga_ref (int, CpoIntVar): FPGA reference clock.
            converter (conv): Converter object.

        Returns:
            dict: Updated configuration dictionary.
        """
        if self.parent.force_cpll:
            v = 1
        else:
            v = [0, 1]
        config[converter.name + "_use_cpll"] = self._convert_input(
            v, converter.name + "_use_cpll"
        )
        if v == [0, 1]:
            self.model.add_kpi(
                config[converter.name + "_use_cpll"],
                name=converter.name + "_use_cpll",
            )
        # Add variables
        config[converter.name + "_m_cpll"] = self._convert_input(
            self._M, converter.name + "_m_cpll"
        )
        # This is documented oddly
        # There are footnotes that include 16 with GTHs, and 16,32 with GTYs
        # but its says "not supported for CPLL"?!?!?
        config[converter.name + "_d_cpll"] = self._convert_input(
            self._D, converter.name + "_d_cpll"
        )
        config[converter.name + "_n1_cpll"] = self._convert_input(
            self._N1, converter.name + "_n1_cpll"
        )
        config[converter.name + "_n2_cpll"] = self._convert_input(
            self._N2, converter.name + "_n2_cpll"
        )

        # Add intermediate variables
        config[converter.name + "_pll_out_cpll"] = self._add_intermediate(
            fpga_ref
            * config[converter.name + "_n1_cpll"]
            * config[converter.name + "_n2_cpll"]
            / (config[converter.name + "_m_cpll"])
        )

        # Add constraints
        self._add_equation(
            [
                if_then(
                    config[converter.name + "_use_cpll"] == 1,
                    config[converter.name + "_d_cpll"] * converter.bit_clock
                    == config[converter.name + "_pll_out_cpll"] * 2,
                ),
            ]
        )
        self._add_equation(
            [
                if_then(
                    config[converter.name + "_use_cpll"] == 1,
                    config[converter.name + "_pll_out_cpll"] >= self.vco_min,
                ),
                if_then(
                    config[converter.name + "_use_cpll"] == 1,
                    config[converter.name + "_pll_out_cpll"] <= self.vco_max,
                ),
            ]
        )

        return config


class QPLL(PLLCommon):
    """QPLL for 7 series FPGAs."""

    M_available = [1, 2, 3, 4]
    _M = [1, 2, 3, 4]

    @property
    def M(self) -> Union[int, List[int]]:
        """Get the M value for the QPLL."""
        return self._M

    @M.setter
    def M(self, value: Union[int, List[int]]) -> None:
        """Set the M value for the QPLL."""
        self._check_in_range(value, self.M_available, "M")
        self._M = value

    N_available = [16, 20, 32, 40, 64, 66, 80, 100]
    _N = [16, 20, 32, 40, 64, 66, 80, 100]

    @property
    def N(self) -> Union[int, List[int]]:
        """Get the N value for the QPLL."""
        return self._N

    @N.setter
    def N(self, value: Union[int, List[int]]) -> None:
        """Set the N value for the QPLL."""
        self._check_in_range(value, self.N_available, "N")
        self._N = value

    D_available = [1, 2, 4, 8, 16]
    _D = [1, 2, 4, 8, 16]

    @property
    def D(self) -> Union[int, List[int]]:
        """Get the D value for the QPLL."""
        return self._D

    @D.setter
    def D(self, value: Union[int, List[int]]) -> None:
        """Set the D value for the QPLL."""
        self._check_in_range(value, self.D_available, "D")
        self._D = value

    @property
    def vco_min(self) -> int:
        """Get the minimum VCO frequency for the transceiver type.

        Returns:
            int: Minimum VCO frequency.

        Raises:
            Exception: Unsupported transceiver type.
        """
        if self.parent.transceiver_type[:3] == "GTH":
            return 8000000000
        elif (
            self.parent.transceiver_type[:3] == "GTX"
        ):  # FIXME: This is only the lower band
            return 5930000000
        raise Exception(
            f"Unknown vco_min for transceiver type {self.parent.transceiver_type}"
        )

    @property
    def vco_max(self) -> int:
        """Get the maximum VCO frequency for the transceiver type.

        Returns:
            int: Maximum VCO frequency.

        Raises:
            Exception: Unsupported transceiver type.
        """
        if self.parent.transceiver_type[:3] == "GTH":
            if float(str(self.parent.speed_grade)[:2]) >= -2:
                return 10312500000
            return 13100000000
        elif (
            self.parent.transceiver_type[:3] == "GTX"
        ):  # FIXME: This is only the lower band
            if float(str(self.parent.speed_grade)[:2]) >= -2:
                return 10312500000
            return 12500000000
        raise Exception(
            f"Unknown vco_max for transceiver type {self.parent.transceiver_type}"
        )

    def add_constraints(
        self, config: dict, fpga_ref: Union[int, CpoIntVar], converter: conv
    ) -> dict:
        """Add constraints for QPLL for 7 series FPGAs.

        Args:
            config (dict): Configuration dictionary.
            fpga_ref (int, CpoIntVar): FPGA reference clock.
            converter (conv): Converter object.

        Returns:
            dict: Updated configuration dictionary.

        Raises:
            Exception: Unsupported transceiver type.
        """
        if self.parent.force_qpll:
            v = 1
        else:
            v = [0, 1]
        # Global flag to use QPLLn
        config[converter.name + "_use_qpll"] = self._convert_input(
            v, converter.name + "_use_qpll"
        )
        if v == [0, 1]:
            self.model.add_kpi(
                config[converter.name + "_use_qpll"],
                name=converter.name + "_use_qpll",
            )

        # Add variables
        config[converter.name + "_m_qpll"] = self._convert_input(
            self._M, converter.name + "_m_qpll"
        )
        config[converter.name + "_d_qpll"] = self._convert_input(
            self._D, converter.name + "_d_qpll"
        )
        config[converter.name + "_n_qpll"] = self._convert_input(
            self._N, converter.name + "_n_qpll"
        )

        # Add intermediate variables
        config[converter.name + "_vco_qpll"] = self._add_intermediate(
            fpga_ref
            * config[converter.name + "_n_qpll"]
            / config[converter.name + "_m_qpll"]
        )
        config[converter.name + "_pll_out_qpll"] = self._add_intermediate(
            fpga_ref
            * config[converter.name + "_n_qpll"]
            / (config[converter.name + "_m_qpll"] * 2)
        )

        # Add constraints
        self._add_equation(
            [
                if_then(
                    config[converter.name + "_use_qpll"] == 1,
                    converter.bit_clock
                    == config[converter.name + "_pll_out_qpll"]
                    * 2
                    / config[converter.name + "_d_qpll"],
                ),
            ]
        )

        if self.parent.transceiver_type[:3] == "GTH":
            self._add_equation(
                [
                    if_then(
                        config[converter.name + "_use_qpll"] == 1,
                        config[converter.name + "_vco_qpll"] >= int(self.vco_min),
                    ),
                    if_then(
                        config[converter.name + "_use_qpll"] == 1,
                        config[converter.name + "_vco_qpll"] <= int(self.vco_max),
                    ),
                ]
            )
        elif self.parent.transceiver_type[:3] == "GTX":
            config[converter.name + "_lower_band_qpll"] = self._convert_input(
                [0, 1], converter.name + "_lower_band_qpll"
            )
            # Lower band
            c1 = if_then(
                config[converter.name + "_use_qpll"] == 1,
                if_then(
                    config[converter.name + "_lower_band_qpll"] == 1,
                    config[converter.name + "_vco_qpll"] >= int(self.vco_min),
                ),
            )
            c2 = if_then(
                config[converter.name + "_use_qpll"] == 1,
                if_then(
                    config[converter.name + "_lower_band_qpll"] == 1,
                    config[converter.name + "_vco_qpll"] <= int(8e9),
                ),
            )
            # See page 72 of https://docs.amd.com/v/u/en-US/ds191-XC7Z030-XC7Z045-data-sheet # noqa: B950
            c5 = if_then(
                config[converter.name + "_use_qpll"] == 1,
                if_then(
                    config[converter.name + "_lower_band_qpll"] == 1,
                    config[converter.name + "_d_qpll"] <= 8,  # no 16
                ),
            )

            # Upper band
            c3 = if_then(
                config[converter.name + "_use_qpll"] == 1,
                if_then(
                    config[converter.name + "_lower_band_qpll"] == 0,
                    config[converter.name + "_vco_qpll"] >= int(9.8e9),
                ),
            )
            c4 = if_then(
                config[converter.name + "_use_qpll"] == 1,
                if_then(
                    config[converter.name + "_lower_band_qpll"] == 0,
                    config[converter.name + "_vco_qpll"] <= int(self.vco_max),
                ),
            )
            self._add_equation([c1, c2, c3, c4, c5])

        else:
            raise Exception("Unsupported transceiver type")

        return config

    def get_config(
        self, config: dict, converter: conv, fpga_ref: Union[int, float]
    ) -> dict:
        """Get QPLL configuration for 7 series FPGAs.

        Args:
            config (dict): Configuration dictionary.
            converter (conv): Converter object.
            fpga_ref (Union[int, float]): FPGA reference clock.

        Returns:
            dict: Updated configuration dictionary.
        """
        pll_config = {"type": "qpll"}
        for k in ["m", "d", "n"]:
            pll_config[k] = self._get_val(config[converter.name + "_" + k + "_qpll"])

        pll_config["vco"] = fpga_ref * pll_config["n"] / pll_config["m"]
        pll_clk_out = fpga_ref * pll_config["n"] / (pll_config["m"] * 2)
        # Check
        assert (
            pll_clk_out * 2 / pll_config["d"] == converter.bit_clock  # type: ignore # noqa: B950
        ), (
            f"Invalid QPLL lane rate {pll_config['vco'] * 2 / pll_config['d']} "
            + f"!= {converter.bit_clock}"
        )

        return pll_config



"""Translation methods for solvers and module."""

from typing import List, Optional, Union

import numpy as np

from adijif.solvers import GK_Intermediate
from adijif.solvers import (
    binary_var,
    CpoExpr,
    CpoFunctionCall,
    CpoSolveResult,
    GEKKO,
    GK_Operators,
    GKVariable,
    integer_var,
)


class gekko_translation:
    """Collection of utility functions to translate to and from solver types."""

    # @property
    # @abstractmethod
    # def model(self):
    #     raise NotImplementedError
    model: Union[GEKKO, CpoExpr] = None

    solution: CpoSolveResult = None

    solver = "gekko"  # "CPLEX"

    def _add_intermediate(
        self, eqs: Union[GK_Operators, CpoFunctionCall]
    ) -> Union[GK_Intermediate, CpoFunctionCall]:
        """Add intermediate/simplified equation.

        Args:
            eqs (GK_Operators, CpoFunctionCall): Equation

        Returns:
            GK_Intermediate/CpoFunctionCall: Converted equation for solver

        Raises:
            Exception: Unknown solver selected
        """
        if self.solver == "gekko":
            return self.model.Intermediate(eqs)
        elif self.solver == "CPLEX":
            return eqs
        else:
            raise Exception(f"Unknown solver: {self.solver}")

    def _add_equation(
        self,
        eqs: List[Union[GKVariable, GK_Intermediate, GK_Operators, CpoExpr]],
    ) -> None:
        """Add equation or relation to solver.

        Args:
            eqs (List[Union[GKVariable, GK_Intermediate, GK_Operators, CpoExpr]]):
                List of equations or CPLEX constraints to add to solver.

        Raises:
            Exception: Unknown solver selected
        """
        if not isinstance(eqs, list):
            eqs = [eqs]

        if self.solver == "gekko":
            self.model.Equations(eqs)
        elif self.solver == "CPLEX":
            for eq in eqs:
                self.model.add_constraint(eq)
        else:
            raise Exception(f"Unknown solver {self.solver}")

    def _get_val(
        self,
        value: Union[int, float, GKVariable, GK_Intermediate, GK_Operators],
    ) -> Union[int, float, str]:
        """Extract value from solver types.

        Args:
            value (GKVariable, GK_Intermediate, GK_Operators): Solver variable

        Returns:
            int/float: Extracted value

        Raises:
            Exception: Unknown solver selected
        """
        if isinstance(value, (float, int)):
            return value
        if self.solver == "gekko":
            if type(value) in [
                GKVariable,
                GK_Intermediate,
            ]:
                return value.value[0]
            elif type(value) is GK_Operators:
                return value.value
            else:
                return value
        elif self.solver == "CPLEX":
            if isinstance(value, (int, float)):
                return value
            return self.solution.get_value(value.get_name())
        else:
            raise Exception(f"Unknown solver {self.solver}")

    def _check_in_range(
        self,
        value: Union[int, str, List[int], List[str]],
        possible: Union[List[int], List[str]],
        varname: str,
    ) -> None:
        """Check if desired value in list.

        Args:
            value (int, str, List[int], List[str]): Desired input value
            possible (List[int], List[str]): Possible valid options
            varname (str): Name of variable

        Raises:
            Exception: Value not indexable from list or possible

        """
        if not isinstance(value, list):
            value = [value]  # type: ignore
        for v in value:  # type: ignore
            if v not in possible:
                raise Exception(f"{v} invalid for {varname}. Only {possible} possible")

    def _convert_input(
        self,
        val: Union[int, List[int], float, List[float]],
        name: Optional[str] = None,
        default: Union[int, float] = None,
    ) -> Union[CpoExpr, GKVariable, GK_Operators]:
        """Convert input to solver variables.

        Args:
            val (int, List[int], float, List[float]): Values or list of
                values to convert to solver variables.
            name (Optional[str]): Name of variable
            default (Optional[int, float]): Default/initial value

        Returns:
            CpoExpr, GKVariable, GK_Operators: Solver variables

        Raises:
            Exception: Unknown solver selected
        """
        if self.solver == "gekko":
            name = None
            return self._convert_input_gekko(val, name, default)
        elif self.solver == "CPLEX":
            return self._convert_input_cplex(val, name)
        else:
            raise Exception(f"Unknown solver {self.solver}")

    def _convert_input_gekko(
        self,
        val: Union[int, List[int], float, List[float]],
        name: Optional[str] = None,
        default: Optional[Union[int, float]] = None,
    ) -> Union[GKVariable, GK_Operators]:
        """Convert input to GEKKO solver variables.

        Args:
            val (int, List[int], float, List[float]): Values or list of
                values to convert to solver variables.
            name (str): Name of variable
            default (Optional[int, float]): Default/initial value

        Returns:
            GKVariable, GK_Operators: Solver variables
        """
        if isinstance(val, list) and len(val) > 1:
            return self._convert_list(val, name, default)
        if name:
            name + "_Const"
        return self.model.Const(value=val, name=name)

    def _convert_input_cplex(
        self,
        val: Union[int, List[int], float, List[float]],
        name: Optional[str] = None,
    ) -> CpoExpr:
        """Convert input to CPLEX solver variables.

        Args:
            val (int, List[int], float, List[float]): Values or list of
                values to convert to solver variables.
            name (str): Name of variable

        Returns:
            CpoExpr: Solver variables

        Raises:
            Exception: Variable already exists in solver model
        """
        if name:
            names = [var.get_name() for var in self.model.get_all_variables()]
            if name in names:
                raise Exception(f"Variable {name} already exists in model")
        if isinstance(val, list) and val.sort() == [0, 1]:
            return binary_var(name=name)
        # if isinstance(val, (float, int)):
        #     return constant(val)
        # return integer_var(domain=(val, val), name=name)
        if isinstance(val, int):
            return val
        if isinstance(val, float):
            if float(int(val)) == val:
                val = int(val)
            return val
            # return integer_var(domain=(val, val), name=name)
            # return self.model.continuous_var(domain=(val, val), name=name)
        return integer_var(domain=val, name=name)

    def _convert_list(
        self,
        val: Union[List[int], List[float]],
        name: Optional[str] = None,
        default: Optional[Union[int, float]] = None,
    ) -> GK_Operators:
        """Convert input list to GEKKO solver variables.

        Args:
            val (List[int], List[float]): List of values to convert
                to solver variables.
            name (str): Name of variable
            default (Optional[int, float]): Default/initial value

        Returns:
            GK_Operators: Solver variables

        Raises:
            Exception: Unsupported case
        """
        # Check if contiguous by simply stride
        delta = val[0] - val[1]
        for i in range(len(val) - 1):
            if val[i] - val[i + 1] is not delta:
                # Must use SOS2
                # print(val[i] - val[i + 1], delta)
                # return self._convert_list2sos(val, name)
                return self.model.sos1(val)

        if np.abs(delta) == 1:  # Easy mode
            # print(np.min(val), np.max(val))
            if not default:
                default = np.min(val)
            if name:
                name + "_Var"
            return self.model.Var(
                integer=True,
                lb=np.min(val),
                ub=np.max(val),
                value=default,
                name=name,
            )

        else:
            # SOS practical in small cases
            if len(val) < 6:
                print("SOS pract")
                # Need to still handle name
                return self.model.sos1(val)
            # Since stride is not zero the best is to use a scale
            # factor and intermediate for best solving performance

            # Need to fit-> (Array+B)*C+D == org_arrar
            # 1,3,5 -> ([1,2,3]-1) * 2 + 1
            # 3,6,9 -> [1,2,3]*3
            # 2,4,6 -> [1,2,3]*2

            raise Exception("NOT COMPLETE")

            # val = np.array(val)
            # vlen = len(val)
            # Array = np.array(range(1, vlen + 1))
            # for B in range(0, vlen):
            #     for C in range(1, vlen):
            #         for D in range(0, vlen):
            #             if val == (Array + B) * C + D:
            #                 break
            # array = self.model.Var(
            #     integer=True, lb=1, ub=4, value=1, name=name + "_Var"
            # )

    # def _convert_back(self, value):
    #     return value



# flake8: noqa

## Broken in 3.8+
# def pytest_exception_interact(node, call, report):
#     excinfo = call.excinfo
#     if "script" in node.funcargs:
#         excinfo.traceback = excinfo.traceback.cut(path=node.funcargs["script"])
#     report.longrepr = node.repr_failure(excinfo)



"""Helper functions for drawing ADC diagrams."""

import logging
from typing import Optional

import adijif as jif

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def draw_adc(adc: Optional[object] = None) -> str:
    """Draw ADC clock tree diagram.

    Args:
        adc: ADC converter object. If None, uses ad9680 as default.

    Returns:
        Path to generated SVG file.
    """
    if adc is None:
        adc = jif.ad9680()

    # Check static
    adc.validate_config()

    required_clocks = adc.get_required_clocks()
    required_clock_names = adc.get_required_clock_names()

    # Add generic clock sources for solver
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):  # noqa: B905
        clk = jif.types.arb_source(name)
        adc._add_equation(clk(adc.model) == clock)
        clks.append(clk)

    # Solve
    solution = adc.model.solve(LogVerbosity="Quiet")
    settings = adc.get_config(solution)

    # Get clock values
    clock_values = {}
    for clk in clks:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    adc.show_rates = False
    return adc.draw(settings["clocks"])


def draw_dac(dac: Optional[object] = None) -> str:
    """Draw DAC clock tree diagram.

    Args:
        dac: DAC converter object. If None, uses ad9144 as default.

    Returns:
        Path to generated SVG file.
    """
    if dac is None:
        dac = jif.ad9144()

    # Check static
    dac.validate_config()

    required_clocks = dac.get_required_clocks()
    required_clock_names = dac.get_required_clock_names()

    # Add generic clock sources for solver
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):  # noqa: B905
        clk = jif.types.arb_source(name)
        dac._add_equation(clk(dac.model) == clock)
        clks.append(clk)

    # Solve
    solution = dac.model.solve(LogVerbosity="Quiet")
    settings = dac.get_config(solution)

    # Get clock values
    clock_values = {}
    for clk in clks:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    dac.show_rates = False
    return dac.draw(settings["clocks"])



"""Cross part connector methods for system models."""

from typing import List, Tuple, Union

import adijif  # noqa: F401
from adijif.clocks.clock import clock as clockc
from adijif.converters.converter import converter as convc
from adijif.fpgas.fpga import fpga as fpgac
from adijif.plls.pll import pll as pllc


class SystemPLL:
    """PLL helper methods for System models."""

    _plls_sysref = []

    @property
    def plls_sysref(self) -> List[pllc]:
        """External PLLs used as SYSREF sources for converters and FPGA.

        Returns:
            List: List of PLL objects
        """
        return self._plls_sysref

    def add_pll_sysref(
        self, pll_name: str, clk: clockc, cnv: convc = None, fpga: fpgac = None
    ) -> None:
        """Add External PLL to system between clock chip, converter and FPGA.

        Args:
            pll_name (str): Name of PLL class
            clk (clockc): Clock chip reference
            cnv (convc): Converter to be driven by PLL
            fpga (fpgac): FPGA to be driven by PLL
        """
        pll = eval(f"adijif.{pll_name}(self.model,solver=self.solver)")  # noqa: S307
        self._plls_sysref.append(pll)
        pll._connected_to_output = []
        pll._ref = clk
        assert cnv or fpga, "Converter or FPGA must be connected to PLL"
        if cnv:
            if hasattr(cnv, "_nested") and cnv._nested:
                # If the converter has nested references (like MxFE)
                # we need to ensure all nested names are connected
                names = cnv._nested
                for name in names:
                    pll._connected_to_output.append(name)
            else:
                pll._connected_to_output.append(cnv.name)
        if fpga:
            pll._connected_to_output.append(fpga.name)

    def _get_ref_clock(
        self, conv: convc, config: dict, clock_names: List[str]
    ) -> Tuple[dict, List[str]]:
        config[f"{conv.name}_ref_clk"] = self.clock._get_clock_constraint(
            f"{conv.name}_ref_clk"
        )
        clock_names.append(f"{conv.name}_ref_clk")
        return config, clock_names

    def _get_ref_clock_fpga(
        self,
        conv: convc,
        config: dict,
        clock_names: List[str],
        need_separate_link_clock: bool,
    ) -> Tuple[dict, List[str]]:
        if conv._nested:
            names = conv._nested
            for name in names:
                config[f"{name}_fpga_ref_clk"] = self.clock._get_clock_constraint(
                    f"{self.fpga.name}_{name}_ref_clk"
                )
                clock_names.append(f"{name}_fpga_ref_clk")
                # sys_refs.append(config[name + "_fpga_ref_clk"])

                if need_separate_link_clock:
                    config[f"{name}_fpga_device_clk"] = (
                        self.clock._get_clock_constraint(
                            f"{self.fpga.name}_{name}_device_clk"
                        )
                    )
                    clock_names.append(f"{name}_fpga_device_clk")

        else:
            config[f"{conv.name}_fpga_ref_clk"] = self.clock._get_clock_constraint(
                f"{self.fpga.name}_{conv.name}_ref_clk"
            )
            clock_names.append(f"{conv.name}_fpga_ref_clk")
            # sys_refs.append(config[f"{conv.name}_fpga_ref_clk"])

            if need_separate_link_clock:
                config[f"{conv.name}_fpga_device_clk"] = (
                    self.clock._get_clock_constraint(
                        f"{self.fpga.name}_{conv.name}_device_clk"
                    )
                )
                clock_names.append(f"{conv.name}_fpga_device_clk")

        return config, clock_names

    def _get_ext_sysref_clock(
        self, name: str, config: dict, clock_names: List[str]
    ) -> Tuple[bool, dict, List[str]]:
        found = False
        if self._plls_sysref:
            for pll in self._plls_sysref:
                if name in pll._connected_to_output:
                    config[f"{name}_sysref"] = pll._get_clock_constraint(
                        f"{name}_sysref"
                    )
                    clock_names.append(f"{name}_sysref")
                    found = True
                    break
        return found, config, clock_names

    def _get_sysref_clock(
        self, conv: convc, config: dict, clock_names: List[str]
    ) -> Tuple[dict, List[str]]:
        if conv._nested:  # MxFE, Transceivers
            assert isinstance(conv._nested, list)
            names = conv._nested
            for name in names:
                # Check if we want to use external PLL for SYSREF
                found, config, clock_names = self._get_ext_sysref_clock(
                    name, config, clock_names
                )
                if not found:
                    # Use clock chip for SYSREF
                    config[f"{name}_sysref"] = self.clock._get_clock_constraint(
                        f"{name}_sysref"
                    )
                    clock_names.append(f"{name}_sysref")
        else:
            found, config, clock_names = self._get_ext_sysref_clock(
                conv.name, config, clock_names
            )
            if not found:
                # Use clock chip for SYSREF
                config[f"{conv.name}_sysref"] = self.clock._get_clock_constraint(
                    f"{conv.name}_sysref"
                )
                clock_names.append(f"{conv.name}_sysref")
        return config, clock_names

    def __apply_ext_sysref_constraint(
        self, name: str, clks: List[str], config: dict, sys_refs: List[str]
    ) -> Tuple[bool, List[str]]:
        found = False
        if self._plls_sysref:
            for pll in self._plls_sysref:
                if name in pll._connected_to_output:
                    self.clock._add_equation(config[f"{name}_sysref"] == clks[1])
                    sys_refs.append(config[f"{name}_sysref"])
                    found = True
                    break
        return found, sys_refs

    def _apply_sysref_constraint(
        self,
        device: Union[convc, fpgac],
        clks: List[str],
        config: str,
        sys_refs: List[str],
    ) -> List[str]:
        if device._nested:
            names = device._nested
            for i, name in enumerate(names):
                found, sys_refs = self.__apply_ext_sysref_constraint(
                    name, clks, config, sys_refs
                )
                if not found:
                    self.clock._add_equation(config[f"{name}_sysref"] == clks[i + 1])
                    sys_refs.append(config[f"{name}_sysref"])
        else:
            found, sys_refs = self.__apply_ext_sysref_constraint(
                device.name, clks, config, sys_refs
            )
            if not found:
                self.clock._add_equation(config[f"{device.name}_sysref"] == clks[1])
                sys_refs.append(config[f"{device.name}_sysref"])

        return sys_refs



"""Collection of utility scripts for specialized checks."""

import copy
from typing import List, Optional

import numpy as np

import adijif.fpgas.xilinx.sevenseries as xp
import adijif.fpgas.xilinx.ultrascaleplus as us
from adijif.converters.converter import converter
from adijif.fpgas.fpga import fpga


def get_jesd_mode_from_params(conv: converter, **kwargs: int) -> List[dict]:
    """Find the JESD mode that matches the supplied parameters.

    Args:
        conv (converter): Converter object of desired device
        kwargs: Parameters and values to match against

    Raises:
        Exception: Converter does not have specified JESD property

    Returns:
        List[dict]: JESD mode that matches the supplied parameters
    """
    results = []
    needed = len(kwargs.items())

    # Remove JESD modes if not needed
    if any([i == "jesd_class" for i in kwargs.items()]):
        modes = conv.quick_configuration_modes[kwargs["jesd_class"]]
        modes = {kwargs["jesd_class"]: modes}
    else:
        modes = conv.quick_configuration_modes

    modes = copy.deepcopy(modes)

    for standard in modes:
        for mode in modes[standard]:
            found = 0
            settings = modes[standard][mode]
            for key, value in kwargs.items():
                if key not in settings:
                    raise Exception(f"{key} not in JESD Configs")
                if isinstance(value, list):
                    for v in value:
                        if settings[key] == v:
                            found += 1
                else:
                    if settings[key] == value:
                        found += 1
            if found == needed:
                results.append(
                    {"mode": mode, "jesd_class": standard, "settings": settings}
                )

    if not results:
        raise Exception(f"No JESD mode found for {kwargs}")

    return results


def get_max_sample_rates(
    conv: converter, fpga: Optional[fpga] = None, limits: Optional[dict] = None
) -> dict:
    """Determine the maximum sample rates for the device.

    Determine the maximum sample rate across all values of M (number of
    virtual converters and supplied limits

    Args:
        conv (converter): Converter object of desired device
        fpga (Optional[fpga]): FPGA object of desired fpga device
        limits (Optional[dict]): Limits to apply to the device and JESD mode

    Raises:
        Exception: Converter does not have specified property
        Exception: Numeric limits must be described in a nested dict
        AttributeError: FPGA object does not have specified property

    Returns:
        dict: Dictionary of maximum sample rates per M
    """
    if fpga:
        max_lanes = fpga.max_serdes_lanes
        if int(fpga.transceiver_type[4]) == 2:
            trx = xp.SevenSeries(parent=fpga, transceiver_type=fpga.transceiver_type)
            max_lane_rate = trx.plls["QPLL"].vco_max
        elif int(fpga.transceiver_type[4]) in [3, 4]:
            trx = us.UltraScalePlus(parent=fpga, transceiver_type=fpga.transceiver_type)
            max_lane_rate = trx.plls["QPLL"].vco_max * 2
        else:
            raise Exception("Unsupported FPGA transceiver type")
    else:
        max_lanes = None
        max_lane_rate = None

    if limits:
        assert isinstance(limits, dict), "limits must be a dictionary"

    results = []
    # Loop across enabled channel counts
    for channels in conv.M_available:
        sample_rates = []
        mode_vals = []
        standards = []
        modes = conv.quick_configuration_modes
        # Cycle through all modes to determine
        for standard in modes:
            for mode in modes[standard]:
                if modes[standard][mode]["M"] not in [channels]:
                    continue
                # Set mode
                conv.set_quick_configuration_mode(mode, standard)
                if max_lanes:
                    if conv.L > max_lanes:
                        continue

                # Set bit_clock
                max_bit_clock = conv.bit_clock_max
                if max_lane_rate:
                    max_bit_clock = min(max_lane_rate, max_bit_clock)

                conv.sample_clock = conv.sample_clock_max
                max_bit_clock_at_max_adc_rate = conv.bit_clock

                # Update model with valid max so we can get true sample clock
                conv.bit_clock = min(
                    max_bit_clock,
                    max_bit_clock_at_max_adc_rate,
                )

                # Apply extra checks
                b = False
                if limits:
                    for limit in limits:
                        if not hasattr(conv, limit):
                            raise AttributeError(
                                f"converter does not have property {limit}"
                            )
                        if isinstance(limits[limit], str):
                            if getattr(conv, limit) != limits[limit]:
                                b = True
                                break
                        elif isinstance(limits[limit], dict):
                            ld = limits[limit]
                            for lc in ld:
                                if lc in [">", "<", "<=", ">=", "=="]:
                                    attr = getattr(conv, limit)
                                    e = eval(  # noqa: S307
                                        f"{attr} {lc} {limits[limit][lc]}"
                                    )
                                    if not e:
                                        b = True
                                        break
                            if b:
                                break

                        else:
                            raise Exception(
                                "Numeric limits must be described in a nested dict"
                            )
                if b:
                    continue

                # Collect sample rate
                sr = min(conv.sample_clock, conv.converter_clock_max)
                sample_rates.append(sr)
                mode_vals.append(mode)
                standards.append(standard)

        if not sample_rates:
            continue

        i = np.argmax(sample_rates)
        mode = mode_vals[i]
        standard = standards[i]
        conv.set_quick_configuration_mode(mode, standard)
        conv.sample_clock = sample_rates[i]
        results.append(
            {
                "sample_clock": conv.sample_clock,
                "bit_clock": conv.bit_clock,
                "L": conv.L,
                "M": conv.M,
                "quick_configuration_mode": mode,
                "jesd_class": standard,
            }
        )
    return results



# flake8: noqa
import pprint

import adijif as jif


def test_adf4030_sysref_example():
    ref_in = int(100e6)

    sync = jif.adf4030()

    output_clocks = [
        2.5e6,
    ]
    output_clocks = list(map(int, output_clocks))  # force to be ints
    clock_names = [
        "SYSREF",
    ]

    sync.set_requested_clocks(ref_in, output_clocks, clock_names)

    sync.solve()

    o = sync.get_config()

    pprint.pprint(o)

    assert o["output_clocks"]["SYSREF"]["rate"] == output_clocks[0]
    assert o["output_clocks"]["SYSREF"]["divider"] == 992


def test_adf4030_system():
    vcxo = 100e6

    import adijif

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.fpga.ref_clock_constraint = "Unconstrained"
    sys.fpga.sys_clk_select = "SYSCLK_QPLL0"  # Use faster QPLL
    sys.fpga.out_clk_select = (
        "XCVR_PROGDIV_CLK"  # force reference to be core clock rate
    )
    cddc = 6
    fddc = 4
    sys.converter.adc.sample_clock = 2900000000 / (cddc * fddc)
    sys.converter.dac.sample_clock = 5800000000 / (cddc * fddc)
    # sys.converter.adc.decimation = 8 * 6
    # sys.converter.dac.interpolation = 4 * 12
    sys.converter.adc.datapath.cddc_decimations = [cddc] * 4
    sys.converter.adc.datapath.fddc_decimations = [fddc] * 8
    sys.converter.adc.datapath.fddc_enabled = [True] * 8
    sys.converter.dac.datapath.cduc_interpolation = cddc
    sys.converter.dac.datapath.fduc_interpolation = fddc
    sys.converter.dac.datapath.fduc_enabled = [True] * 8
    assert sys.converter.dac.interpolation == cddc * fddc

    # sys.converter.clocking_option = "direct"
    # sys.add_pll_inline("adf4371", sys.clock, sys.converter)
    sys.add_pll_sysref("adf4030", vcxo, sys.converter, sys.fpga)

    mode_tx = "0"
    mode_rx = "1.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204c")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204c")

    assert sys.converter.adc.M == 8
    assert sys.converter.adc.F == 12
    assert sys.converter.adc.K == 64
    assert sys.converter.adc.Np == 12
    assert sys.converter.adc.CS == 0
    assert sys.converter.adc.L == 1
    assert sys.converter.adc.S == 1

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    cfg = sys.solve()

    pprint.pprint(cfg)
    print(sys.converter.dac.converter_clock)


def test_adf4030_chained_clocks():
    import adijif

    L = 2  # max_available_lanes_per_ad9084_side
    M = 4  # Converters used to AD9084 side
    Np = 16  # Bits per sample

    vcxo = int(400e6)
    cddc_dec = 4
    fddc_dec = 8
    converter_rate = int(12.8e9)

    sys = adijif.system("ad9084_rx", "ltc6952", "xilinx", vcxo, solver="CPLEX")

    sys.fpga.setup_by_dev_kit_name("vcu118")
    sys.converter.sample_clock = converter_rate / (cddc_dec * fddc_dec)
    sys.converter.datapath.cddc_decimations = [cddc_dec] * 4
    sys.converter.datapath.fddc_decimations = [fddc_dec] * 8
    sys.converter.datapath.fddc_enabled = [True] * 8

    sys.converter.clocking_option = "direct"
    sys.add_pll_inline("adf4382", vcxo, sys.converter)
    # IMPLEMENTATION HERE
    sys.add_pll_sysref("adf4030", sys.clock, sys.converter, sys.fpga)

    sys.clock.minimize_feedback_dividers = False

    mode_rx = adijif.utils.get_jesd_mode_from_params(
        sys.converter, M=M, L=L, Np=Np, jesd_class="jesd204c"
    )
    print(f"RX JESD Mode: {mode_rx}")
    assert mode_rx
    mode_rx = mode_rx[0]["mode"]

    sys.converter.set_quick_configuration_mode(mode_rx, "jesd204c")

    print(f"Lane rate: {sys.converter.bit_clock/1e9} Gbps")
    print(f"Needed Core clock: {sys.converter.bit_clock/66} MHz")

    assert sys.converter.bit_clock == int(13.2e9)

    cfg = sys.solve()

    pprint.pprint(cfg)
    # 'clock_ext_pll_sysref_adf4030': {'n': 155,
    #                               'out_dividers': [4030],
    #                               'output_clocks': {'AD9084_RX_sysref': {'divider': 4030,
    #                                                                      'rate': 591715.976331361}},
    #                               'r': 13,
    #                               'vco': 2384615384.6153846},

    # Verify sysref is within LMFC
    mfc = sys.converter.multiframe_clock
    estimate = cfg["clock_ext_pll_sysref_adf4030"]["output_clocks"]["AD9084_RX_sysref"][
        "rate"
    ]
    for lmfc_div in range(1, 21):
        sysref_rate = mfc / lmfc_div
        if sysref_rate == estimate:
            print(f"Found sysref match: {sysref_rate} with divisor {lmfc_div}")
            return


def test_adf4030_with_arb_source():
    """Test ADF4030 with arb_source input_ref."""
    input_ref = jif.types.arb_source(
        "ref_in", a_min=90000000, a_max=110000000, b_min=1, b_max=1
    )

    sync = jif.adf4030(solver="CPLEX")

    output_clocks = [2.5e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["SYSREF"]

    sync.set_requested_clocks(input_ref, output_clocks, clock_names)

    sync.solve()

    o = sync.get_config()

    assert o["output_clocks"]["SYSREF"]["rate"] == output_clocks[0]
    assert isinstance(o["vco"], (int, float))


def test_adf4030_with_range():
    """Test ADF4030 with range input_ref."""
    input_ref = jif.types.range(90000000, 110000000, 5000000, "ref_in")

    sync = jif.adf4030(solver="CPLEX")

    output_clocks = [2.5e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["SYSREF"]

    sync.set_requested_clocks(input_ref, output_clocks, clock_names)

    sync.solve()

    o = sync.get_config()

    assert o["output_clocks"]["SYSREF"]["rate"] == output_clocks[0]
    assert isinstance(o["vco"], (int, float))



# flake8: noqa
from pprint import pprint

import pytest

import adijif as jif
import adijif.fpgas.xilinx.sevenseries as xp
import adijif.fpgas.xilinx.ultrascaleplus as us


@pytest.mark.parametrize("pll, out_clock", [("cpll", 0.5e9), ("qpll", 10e9)])
def test_7s_pll(pll, out_clock):
    ss_plls = xp.SevenSeries(transceiver_type="GTXE2", solver="CPLEX")

    in_clock = int(100e6)
    out_clock = int(out_clock)
    cnv = jif.ad9680()

    # Change sample rate to get to the desired bit clock
    # sample_rate * cnv.Np * cnv.M * (cnv.encoding_d / cnv.encoding_n) / cnv.L = bit_clock
    sample_rate = (
        out_clock * cnv.L / (cnv.Np * cnv.M * (cnv.encoding_d / cnv.encoding_n))
    )
    cnv.sample_clock = sample_rate

    assert cnv.bit_clock == out_clock

    cfg = {}
    config = ss_plls.add_constraints(cfg, in_clock, cnv)

    print(ss_plls.model.export_model())

    ss_plls.solve()
    print(ss_plls.solution.print_solution())

    cfg = ss_plls.get_config(config, cnv, in_clock)
    pprint(cfg)

    if cfg["type"] == "cpll":
        pll_out = (in_clock * cfg["n1"] * cfg["n2"]) / cfg["m"]
        lane_rate = pll_out * 2 / cfg["d"]
    else:
        pll_out = in_clock * cfg["n"] / (cfg["m"] * 2)
        lane_rate = pll_out * 2 / cfg["d"]

    assert lane_rate == cnv.bit_clock

    assert cfg["type"] == pll


@pytest.mark.parametrize(
    "pll, out_clock",
    [
        ("cpll", 0.5e9),
        ("qpll", 14e9),
        ("qpll", 32e9),
        ("qpll1", 9e9),
    ],  # No frac mode
)
def test_us_pll(pll, out_clock):
    us_plls = us.UltraScalePlus(transceiver_type="GTYE4", solver="CPLEX")
    us_plls.plls["QPLL"].force_integer_mode = True
    us_plls.plls["QPLL1"].force_integer_mode = True

    in_clock = int(100e6)
    out_clock = int(out_clock)
    cnv = jif.ad9680()

    # Change sample rate to get to the desired bit clock
    # sample_rate * cnv.Np * cnv.M * (cnv.encoding_d / cnv.encoding_n) / cnv.L = bit_clock
    sample_rate = (
        out_clock * cnv.L / (cnv.Np * cnv.M * (cnv.encoding_d / cnv.encoding_n))
    )
    cnv.sample_clock = sample_rate

    assert cnv.bit_clock == out_clock

    cfg = {}
    config = us_plls.add_constraints(cfg, in_clock, cnv)

    print(us_plls.model.export_model())

    us_plls.solve()
    print(us_plls.solution.print_solution())

    cfg = us_plls.get_config(config, cnv, in_clock)
    pprint(cfg)

    if cfg["type"] == "cpll":
        pll_out = (in_clock * cfg["n1"] * cfg["n2"]) / cfg["m"]
        lane_rate = pll_out * 2 / cfg["d"]
    else:
        pll_out = in_clock * cfg["n"] / (cfg["m"] * cfg["clkout_rate"])
        lane_rate = pll_out * 2 / cfg["d"]

    assert lane_rate == cnv.bit_clock

    assert cfg["type"] == pll



"""CLI entry point for PyADI-JIF Tools Explorer."""

import os
import sys
from typing import Optional

from streamlit.web import cli as stcli


def run_streamlit(args: Optional[list[str]] = None) -> None:
    """Run the Streamlit tools explorer app."""
    # Get the directory where this file is located
    current_dir = os.path.dirname(os.path.abspath(__file__))
    main_file = os.path.join(current_dir, "main.py")

    # Get arguments from sys.argv if not provided
    if args is None:
        args = sys.argv[1:]

    # Set up sys.argv for streamlit
    sys.argv = [
        "streamlit",
        "run",
        main_file,
        "--logger.level=error",
        "--browser.gatherUsageStats=false",
        "--server.showEmailPrompt=false",
    ] + args

    # Run streamlit
    sys.exit(stcli.main())


if __name__ == "__main__":
    args = sys.argv[1:]
    run_streamlit(args)



# flake8: noqa
import pprint

import pytest

import adijif as jif


def test_get_jesd_mode_from_params_with_jesd_class():
    """Test get_jesd_mode_from_params with jesd_class specified."""
    conv = jif.ad9081_rx()

    # Test with jesd_class filter
    results = jif.utils.get_jesd_mode_from_params(conv, jesd_class="jesd204b", M=4)
    assert len(results) > 0
    for result in results:
        assert result["jesd_class"] == "jesd204b"
        assert result["settings"]["M"] == 4


def test_get_jesd_mode_from_params_with_list_values():
    """Test get_jesd_mode_from_params with list of values."""
    conv = jif.ad9081_rx()

    # Test with list of values
    results = jif.utils.get_jesd_mode_from_params(conv, M=[2, 4])
    assert len(results) > 0
    for result in results:
        assert result["settings"]["M"] in [2, 4]


def test_get_jesd_mode_from_params_invalid_key():
    """Test get_jesd_mode_from_params with invalid key."""
    conv = jif.ad9081_rx()

    # Test with invalid key should raise exception
    with pytest.raises(Exception, match="not in JESD Configs"):
        jif.utils.get_jesd_mode_from_params(conv, invalid_key=123)


def test_get_jesd_mode_from_params_no_match():
    """Test get_jesd_mode_from_params when no mode matches."""
    conv = jif.ad9081_rx()

    # Test with impossible parameters should raise exception
    with pytest.raises(Exception, match="No JESD mode found"):
        jif.utils.get_jesd_mode_from_params(conv, M=999)


def test_generate_max_rates_fpga_limits_utility():
    """Test get_max_sample_rates with FPGA and various limits."""
    conv = jif.ad9081_rx()

    # Determine max data rate from FPGA
    fpga = jif.xilinx()
    fpga.setup_by_dev_kit_name("zc706")
    fpga.sys_clk_select = "XCVR_QPLL0"  # Use faster QPLL

    limits = {"jesd_class": "jesd204b", "bit_clock": {"<=": 10e9}}

    results = jif.utils.get_max_sample_rates(conv, fpga, limits)

    assert len(results) > 0
    for result in results:
        assert result["bit_clock"] <= 10e9
        assert result["jesd_class"] == "jesd204b"


def test_generate_max_rates_with_numeric_limits():
    """Test get_max_sample_rates with numeric comparison limits."""
    conv = jif.ad9081_rx()

    # Test with greater than limit
    limits = {"L": {">": 3}}
    results = jif.utils.get_max_sample_rates(conv, limits=limits)

    assert len(results) > 0
    for result in results:
        assert result["L"] > 3


def test_generate_max_rates_with_invalid_limit_format():
    """Test get_max_sample_rates with invalid limit format."""
    conv = jif.ad9081_rx()

    # Test with invalid limit format (not dict or string)
    limits = {"L": [1, 2, 3]}  # list is invalid

    with pytest.raises(
        Exception, match="Numeric limits must be described in a nested dict"
    ):
        jif.utils.get_max_sample_rates(conv, limits=limits)


def test_generate_max_rates_with_invalid_property():
    """Test get_max_sample_rates with invalid converter property."""
    conv = jif.ad9081_rx()

    # Test with property that doesn't exist
    limits = {"nonexistent_property": {"==": 10}}

    with pytest.raises(AttributeError, match="does not have property"):
        jif.utils.get_max_sample_rates(conv, limits=limits)


def test_generate_max_rates_with_string_limits():
    """Test get_max_sample_rates with string value limits."""
    conv = jif.ad9680()

    # Test with string comparison
    limits = {"jesd_class": "jesd204b"}
    results = jif.utils.get_max_sample_rates(conv, limits=limits)

    assert len(results) > 0
    for result in results:
        assert result["jesd_class"] == "jesd204b"


def test_generate_max_rates_unsupported_fpga():
    """Test get_max_sample_rates with unsupported FPGA transceiver type."""
    conv = jif.ad9680()

    fpga = jif.xilinx()
    fpga.setup_by_dev_kit_name("zc706")
    # Mock an unsupported transceiver type
    fpga.transceiver_type = "GTH19"  # Invalid - position 4 should be 2, 3, or 4

    with pytest.raises(Exception, match="Unsupported FPGA transceiver type"):
        jif.utils.get_max_sample_rates(conv, fpga=fpga)


def test_generate_max_rates_with_ultrascaleplus():
    """Test get_max_sample_rates with UltraScale+ FPGA."""
    conv = jif.ad9680()

    fpga = jif.xilinx()
    fpga.setup_by_dev_kit_name("vcu118")  # UltraScale+ board

    results = jif.utils.get_max_sample_rates(
        conv, fpga=fpga, limits={"jesd_class": "jesd204b"}
    )

    assert len(results) > 0


# def test_generate_max_rates_fpga_limits_utility():
#     conv = jif.ad9081_rx()

#     # Determine max data rate from FPGA
#     fpga = jif.xilinx()
#     fpga.setup_by_dev_kit_name("zc706")
#     fpga.sys_clk_select = "GTH34_SYSCLK_QPLL0"  # Use faster QPLL

#     limits = {"jesd_class": "jesd204b", "bit_clock": {"<=": 10e9}}

#     results = jif.utils.get_max_sample_rates(conv, fpga, limits)

#     ref = []
#     assert results == ref


def test_generate_max_rates_fpga_utility():
    conv = jif.ad9081_rx()
    # conv.decimation = 1

    # Determine max data rate from FPGA
    fpga = jif.xilinx()
    fpga.setup_by_dev_kit_name("zc706")
    fpga.sys_clk_select = "XCVR_QPLL0"  # Use faster QPLL

    limits = {"jesd_class": "jesd204b"}

    results = jif.utils.get_max_sample_rates(conv, fpga, limits)

    ref = [
        {
            "sample_clock": 4000000000.0,
            "bit_clock": 10000000000.0,
            "L": 8,
            "M": 1,
            "quick_configuration_mode": "19.01",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 2062500000.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 2,
            "quick_configuration_mode": "19.0",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 515625000.0,
            "bit_clock": 10312500000.0,
            "L": 3,
            "M": 3,
            "quick_configuration_mode": "9.01",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 1031250000.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 4,
            "quick_configuration_mode": "18.0",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 515625000.0,
            "bit_clock": 10312500000.0,
            "L": 6,
            "M": 6,
            "quick_configuration_mode": "15.01",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 515625000.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 8,
            "quick_configuration_mode": "16.0",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 257812500.0,
            "bit_clock": 10312500000.0,
            "L": 6,
            "M": 12,
            "quick_configuration_mode": "15.1",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 257812500.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 16,
            "quick_configuration_mode": "17.0",
            "jesd_class": "jesd204b",
        },
    ]
    assert len(results) == len(ref)
    for result in results:
        # print(result)
        if result not in ref:
            pprint.pprint(result)
        assert result in ref


def test_generate_max_rates_utility():
    conv = jif.ad9081_rx()
    # conv.decimation = 1

    results = jif.utils.get_max_sample_rates(conv)

    ref = [
        {
            "L": 8,
            "M": 1,
            "bit_clock": 10000000000.0,
            "quick_configuration_mode": "19.01",
            "sample_clock": 4000000000.0,
            "jesd_class": "jesd204b",
        },
        {
            "L": 4,
            "M": 2,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "14.0",
            "sample_clock": 4000000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 3,
            "M": 3,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "9.01",
            "sample_clock": 1500000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 8,
            "M": 4,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "27.0",
            "sample_clock": 4000000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 6,
            "M": 6,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "15.01",
            "sample_clock": 1500000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 8,
            "M": 8,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "26.0",
            "sample_clock": 2000000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 6,
            "M": 12,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "15.0",
            "sample_clock": 750000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 8,
            "M": 16,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "17.0",
            "sample_clock": 750000000.0,
            "jesd_class": "jesd204c",
        },
    ]

    # pprint.pprint(results)

    assert len(results) == len(ref)
    for result in results:
        if result not in ref:
            pprint.pprint(result)
        assert result in ref



"""Drawing features for AD9680."""

from typing import Dict

from adijif.draw import Layout, Node  # type: ignore # isort: skip  # noqa: I202


class ad9680_draw:
    """AD9680 drawing features."""

    _system_draw = False
    show_rates = True

    def _init_diagram(self) -> None:
        """Initialize diagram for AD9680 alone."""
        self.ic_diagram_node = None
        self._diagram_output_dividers = []

        self.ic_diagram_node = Node("AD9680")

        # External
        # ref_in = Node("REF_IN", ntype="input")
        # lo.add_node(ref_in)

        crossbar = Node("Crossbar", ntype="crossbar")
        self.ic_diagram_node.add_child(crossbar)
        for adc in range(2):
            adc_node = Node(f"ADC{adc}", ntype="adc")
            self.ic_diagram_node.add_child(adc_node)
            adc_node.shape = "parallelogram"
            self.ic_diagram_node.add_connection(
                {"from": adc_node, "to": crossbar, "type": "data"}
            )

        for ddc in range(4):
            ddc_node = Node(f"DDC{ddc}", ntype="ddc")
            self.ic_diagram_node.add_child(ddc_node)
            self.ic_diagram_node.add_connection(
                {"from": crossbar, "to": ddc_node, "type": "data"}
            )

        jesd204_framer = Node("JESD204 Framer", ntype="jesd204framer")
        self.ic_diagram_node.add_child(jesd204_framer)

        for ddc in range(4):
            ddc = self.ic_diagram_node.get_child(f"DDC{ddc}")
            self.ic_diagram_node.add_connection(
                {"from": ddc, "to": jesd204_framer, "type": "data"}
            )

    def _update_diagram(self, config: Dict) -> None:
        """Update diagram with configuration.

        Args:
            config (Dict): Configuration dictionary

        Raises:
            Exception: If key is not D followed by a number
        """
        # Add output dividers
        keys = config.keys()
        output_dividers = self.ic_diagram_node.get_child("Output Dividers")
        for key in keys:
            if key.startswith("D"):
                div = Node(key, ntype="divider")
                output_dividers.add_child(div)
                self.ic_diagram_node.add_connection(
                    {"from": output_dividers, "to": div}
                )
            else:
                raise Exception(
                    f"Unknown key {key}. Must be of for DX where X is a number"
                )

    def draw(
        self, clocks: Dict, lo: Layout = None, clock_chip_node: Node = None
    ) -> str:
        """Draw diagram in d2 language for IC alone with reference clock.

        Args:
            clocks (Dict): Dictionary of clocks
            lo (Layout): Layout object to add to. Defaults to None.
            clock_chip_node (Node): Node to connect to. Defaults to None.

        Returns:
            str: Diagram in d2 language

        Raises:
            Exception: If no solution is saved
        """
        if not self._saved_solution:
            raise Exception("No solution to draw. Must call solve first.")

        system_draw = lo is not None

        if not system_draw:
            lo = Layout("AD9680 Example")
            lo.show_rates = self.show_rates
        else:
            # Verify lo is a Layout object
            assert isinstance(lo, Layout), "lo must be a Layout object"
        lo.add_node(self.ic_diagram_node)

        static_options = self.get_config()

        if not system_draw:
            ref_in = Node("REF_IN", ntype="input")
            lo.add_node(ref_in)
        else:
            to_node = lo.get_node("AD9680_ref_clk")
            # Locate node connected to this one
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            ref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

        for i in range(2):
            adc = self.ic_diagram_node.get_child(f"ADC{i}")
            lo.add_connection(
                {"from": ref_in, "to": adc, "rate": clocks["AD9680_ref_clk"]}
            )

        # Update Node values
        for ddc in range(4):
            rate = clocks["AD9680_ref_clk"]
            self.ic_diagram_node.update_connection("Crossbar", f"DDC{ddc}", rate)

            ddc_node = self.ic_diagram_node.get_child(f"DDC{ddc}")
            ddc_node.value = str(static_options["decimation"])
            drate = rate / static_options["decimation"]

            self.ic_diagram_node.update_connection(f"DDC{ddc}", "JESD204 Framer", drate)

        # Connect clock to framer
        if not system_draw:
            sysref_in = Node("SYSREF_IN", ntype="input")

            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": self.ic_diagram_node.get_child("JESD204 Framer"),
                    "rate": clocks["AD9680_sysref"],
                }
            )
        else:
            to_node = lo.get_node("AD9680_sysref")
            # Locate node connected to this one
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            sysref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": self.ic_diagram_node.get_child("JESD204 Framer"),
                    "rate": clocks["AD9680_sysref"],
                }
            )

        # Connect Remote Deframer
        remote_deframer = Node("JESD204 Deframer", ntype="deframer")
        lo.add_node(remote_deframer)

        # Add connect for each lane
        for _ in range(self.L):
            lane_rate = self.bit_clock
            lo.add_connection(
                {
                    "from": self.ic_diagram_node.get_child("JESD204 Framer"),
                    "to": remote_deframer,
                    "rate": lane_rate,
                    "type": "data",
                }
            )

        if not system_draw:
            return lo.draw()



"""Converter base meta class for all converter clocking models."""

from abc import ABCMeta, abstractmethod
from typing import List, Union

from adijif.converters.converter import converter


class dac(converter, metaclass=ABCMeta):
    """DAC base meta class used to enforce all DAC classes.

    This class should be inherited from for all DACs and transceivers.

    """

    def _check_valid_internal_configuration(self) -> None:
        """Verify current internal clocking configuration for part is valid.

        Raises:
            Exception: Invalid clocking configuration
        """
        if self.interpolation * self.sample_clock > self.converter_clock_max:
            raise Exception(
                "DAC rate too fast for configuration {} (max: {})".format(
                    self.interpolation * self.sample_clock,
                    self.converter_clock_max,
                )
            )
        if self.interpolation * self.sample_clock < self.converter_clock_min:
            raise Exception(
                "DAC rate too slow for configuration {} (min: {})".format(
                    self.interpolation * self.sample_clock,
                    self.converter_clock_min,
                )
            )

    @property
    @abstractmethod
    def interpolation_available(self) -> List:
        """Interpolation settings available.

        Must be a list or None

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    _interpolation = 1

    @property
    def interpolation(self) -> Union[int, float]:
        """Interpolation between DAC and JESD framer. If none device is not an DAC.

        Generally a multiple of 2

        Returns:
            int: interpolation value
        """
        return self._interpolation

    @interpolation.setter
    def interpolation(self, value: int) -> None:
        """Interpolation between DAC and JESD framer. If none device is not an DAC.

        Args:
            value (int): interpolation

        Raises:
            Exception: interpolation not an integer or not in range
        """
        if int(value) != value:
            raise Exception("interpolation_available must be an integer")
        if value not in self.interpolation_available:
            raise Exception("interpolation_available not in range for device")
        self._interpolation = value

    @property
    def converter_clock(self) -> Union[int, float]:
        """Get rate of converter in samples per second.

        Returns:
            float: converter clock rate in samples per second
        """
        return self.sample_clock * self.interpolation
